import csv
import io
import json
from typing import Union, List

from ..coworks import TechMicroService


class CSVMicroService(TechMicroService):

    def post_format(self, content: str = "", title: Union[bool, List[str]] = True,
                    remove_rows: List[int] = None, remove_columns: List[int] = None,
                    delimiter: str = ','):
        """Format JSON list to CSV content."""

        remove_rows = remove_rows if remove_rows is not None else []
        remove_columns = remove_columns if remove_columns is not None else []
        if type(remove_rows) is not list:
            raise EnvironmentError("remove_rows parameter must be a list")
        if type(remove_columns) is not list:
            raise EnvironmentError("remove_columns parameter must be a list")

        rows = json.loads(content)
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC, delimiter=delimiter)

        if title is True:  # takes title from keys of first line
            line = [key for idx, key in enumerate(rows[0]) if idx not in remove_columns]
            writer.writerow(line)
        elif type(title) is list:
            line = [col for idx, col in enumerate(title) if idx not in remove_columns]
            writer.writerow(line)
        for index, row in enumerate(rows):
            if index not in remove_rows:
                line = [col for idx, col in enumerate(row.values()) if idx not in remove_columns]
                writer.writerow(line)

        return output.getvalue()



import functools
import io
import re
import pytest

import yaml

from coworks.cws import TerraformWriter
from coworks.cws.writer.writer import ListWriter, OpenApiWriter
from tests.src.coworks.tech_ms import SimpleMS


def test_export_list():
    simple = SimpleMS()
    writer = ListWriter(simple)
    simple.extensions['writers']['list'].export(output='/dev/null')
    output = io.StringIO()
    writer.export(output=output)
    output.seek(0)
    assert len(output.read()) == 83


def test_export_open_api():
    simple = SimpleMS()
    writer = OpenApiWriter(simple)
    simple.extensions['writers']['openapi'].export(output='/dev/null')
    output = io.StringIO()
    writer.export(output=output)
    output.seek(0)
    print(output.read())
    output.seek(0)
    data = yaml.load(output)
    assert {'info', 'openapi', 'paths', 'version'} == set(data.keys())
    assert {'/', '/content', '/content/{_0}', '/content/{_0}/{_1}', '/extended/content', '/kwparam1',
            '/kwparam2'} == set(data['paths'].keys())
    print(writer.pathes)
    assert len(writer.pathes) == 7
    sum = lambda x, y: x + y
    print([e.keys() for e in writer.pathes.values()])
    assert functools.reduce(sum, [len(e.keys()) for e in writer.pathes.values()], 0) == 9


def test_export_terraform():
    simple = SimpleMS()
    writer = TerraformWriter(simple)
    simple.extensions['writers']['terraform'].export(output='/dev/null')
    output = io.StringIO()
    writer.export(output=output)
    output.seek(0)
    print(output.read())
    output.seek(0)
    assert len(re.sub(r"\s", "", output.read())) == 18226
    print(writer.entries)
    assert len(writer.entries) == 8

    assert writer.entries['_'].parent_uid is None
    assert writer.entries['_'].is_root
    assert writer.entries['_'].path is None
    assert 'GET' in writer.entries['_'].methods
    assert 'POST' not in writer.entries['_'].methods

    assert writer.entries['__content'].parent_uid == '_'
    assert not writer.entries['__content'].is_root
    assert writer.entries['__content'].parent_is_root
    assert writer.entries['__content'].path == 'content'
    assert 'GET' in writer.entries['__content'].methods
    assert 'POST' in writer.entries['__content'].methods

    assert writer.entries['__extended'].methods is None

    assert not writer.entries['__extended_content'].is_root
    assert not writer.entries['__extended_content'].parent_is_root
    assert 'GET' in writer.entries['__extended_content'].methods




import cgi
import io
import json
import os
import sys
import traceback
from functools import update_wrapper

import boto3
from botocore.exceptions import BotoCoreError
from chalice import BadRequestError, Response
from requests_toolbelt.multipart import MultipartDecoder

from coworks.utils import begin_xray_subsegment, end_xray_subsegment


class CoworksMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.debug = kwargs.pop('debug', False)

        self.aws_s3_sfn_data_session = AwsS3Session(env_var_access_key="AWS_RUN_ACCESS_KEY_ID",
                                                    env_var_secret_key="AWS_RUN_SECRET_KEY",
                                                    env_var_region="AWS_RUN_REGION")
        self.aws_s3_form_data_session = AwsS3Session(env_var_access_key="AWS_FORM_DATA_ACCESS_KEY_ID",
                                                     env_var_secret_key="AWS_FORM_DATA_SECRET_KEY",
                                                     env_var_region="AWS_FORM_DATA_REGION")

    def _create_rest_proxy(self, func, kwarg_keys, args, varkw):
        original_app_class = self.__class__

        def proxy(**kws):
            if self.debug:
                print(f"Calling {func} for {self}")

            subsegment = begin_xray_subsegment(f"{func.__name__} microservice")
            try:
                if subsegment:
                    subsegment.put_metadata('headers', self.current_request.headers, "CoWorks")

                # Renames positionnal parameters (index added in label)
                kwargs = {}
                for kw, value in kws.items():
                    param = args[int(kw[1:])]
                    kwargs[param] = value

                # Adds kwargs parameters
                def check_param_expected_in_lambda(param_name):
                    """Alerts when more parameters than expected are defined in request."""
                    if param_name not in kwarg_keys and varkw is None:
                        raise BadRequestError(f"TypeError: got an unexpected keyword argument '{param_name}'")

                def add_param(param_name, param_value):
                    check_param_expected_in_lambda(param_name)
                    if param_name in params:
                        if isinstance(params[param_name], list):
                            params[param_name].append(param_value)
                        else:
                            params[param_name] = [params[param_name], param_value]
                    else:
                        params[param_name] = param_value

                req = self.current_request
                if kwarg_keys or varkw:
                    params = {}
                    if req.raw_body:  # POST request
                        content_type = req.headers['content-type']
                        if content_type.startswith('multipart/form-data'):
                            try:
                                multipart_decoder = MultipartDecoder(req.raw_body, content_type)
                                for part in multipart_decoder.parts:
                                    name, content = self._get_multipart_content(part)
                                    add_param(name, content)
                            except Exception as e:
                                print(e)
                            kwargs = dict(**kwargs, **params)

                        elif content_type.startswith('application/json'):
                            if hasattr(req.json_body, 'items'):
                                params = {}
                                for k, v in req.json_body.items():
                                    add_param(k, v)
                                kwargs = dict(**kwargs, **params)
                            else:
                                kwargs[kwarg_keys[0]] = req.json_body
                        elif content_type.startswith('text/plain'):
                            kwargs[kwarg_keys[0]] = req.json_body
                        else:
                            BadRequestError(f"Cannot manage content type {content_type} for {self}")

                    else:  # GET request

                        # adds parameters from qurey parameters
                        for k in req.query_params or []:
                            value = req.query_params.getlist(k)
                            add_param(k, value if len(value) > 1 else value[0])
                        kwargs = dict(**kwargs, **params)

                if self.debug:
                    print(f"Calling {self} with event {kwargs}")

                # chalice is changing class for local server for threading reason (why not mixin..?)
                self_class = self.__class__
                if self_class != original_app_class:
                    self.__class__ = original_app_class
                resp = func(self, **kwargs)
                self.__class__ = self_class

                return _convert_response(resp)

            except Exception as e:
                print(f"Exception : {str(e)}")
                traceback.print_exc()
                if subsegment:
                    subsegment.add_exception(e, traceback.extract_stack())
                raise BadRequestError(str(e))
            finally:
                end_xray_subsegment()

        proxy = update_wrapper(proxy, func)
        proxy.__class_func__ = func
        return proxy

    def _get_multipart_content(self, part):
        headers = {k.decode('utf-8'): cgi.parse_header(v.decode('utf-8')) for k, v in part.headers.items()}
        content = part.content
        _, content_disposition_params = headers['Content-Disposition']
        part_content_type, _ = headers.get('Content-Type', (None, None))
        name = content_disposition_params['name']

        # content in a text or json value
        if 'filename' not in content_disposition_params:
            if part_content_type == 'application/json':
                return name, self._get_data_on_s3(json.loads(content.decode('utf-8')))
            return name, self._get_data_on_s3(content.decode('utf-8'))

        # content in a file (s3 or plain text)
        if part_content_type == 'text/s3':
            pathes = content.decode('utf-8').split('/', 1)
            try:
                s3_object = self.aws_s3_form_data_session.client.get_object(Bucket=pathes[0], Key=pathes[1])
            except BotoCoreError:
                print(f"Bucket={pathes[0]} Key={pathes[1]} not found on s3")
                return BadRequestError(f"Bucket={pathes[0]} Key={pathes[1]} not found on s3")
            file = io.BytesIO(s3_object['Body'].read())
            mime_type = s3_object['ContentType']
        else:
            file = io.BytesIO(content)
            mime_type = part_content_type
        file.name = content_disposition_params['filename']

        return name, FileParam(file, mime_type)

    def _set_multipart_content(self, form_data):
        def encode_part(_part):
            if type(_part) is str:
                return None, _part, 'text/plain'

            if 'mime_type' in _part:
                mime_type = _part.get('mime_type')
            elif 'json' in _part:
                _part['content'] = _part.get('json')
                mime_type = 'application/json'
            elif 's3' in _part:
                path = _part.get('s3')
                _part['filename'] = path.split('/')[-1]
                _part['path'] = path
                mime_type = 'text/s3'
            else:
                mime_type = 'text/plain'

            filename = _part.get('filename')
            if mime_type == 'text/plain':
                content = _part.get('content')
                return filename, content, mime_type
            elif mime_type == 'application/json':
                content = _part.get('content')
                return filename, json.dumps(content), mime_type
            elif mime_type == 'text/s3':
                path = _part.get('path')
                return filename, path, mime_type
            else:
                raise BadRequestError(f"Undefined mime type {mime_type}")

        parts = []
        for name, part in form_data.items():
            if type(part) is list:
                parts.extend([(name, encode_part(p)) for p in part])
            else:
                parts.append((name, encode_part(part)))
        return parts

    def _set_data_on_s3(self, data):
        """Saves value on S3 temporary file if content is too big."""

        def set_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            context = self.lambda_context
            key = f"tmp/{context.aws_request_id}"
            tags = f"Name={context.function_name}"
            s3_client.put_object(Bucket="coworks-microservice", Key=key, Body=value, Tagging=tags)
            return f"$${key}$$"

        if type(data) == str:
            return set_on_s3(data) if len(data) > 1000 else data
        if type(data) == list:
            return [self._set_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._set_data_on_s3(v)
        return data

    def _get_data_on_s3(self, data):
        """Retrieves value from S3 temporary file (content too big)."""

        def get_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            s3_object = s3_client.get_object(Bucket="coworks-microservice", Key=value[2:-2])
            return s3_object['Body'].read().decode("utf-8")

        if type(data) == str:
            return get_on_s3(data) if data.startswith('$$') and data.endswith('$$') else data
        if type(data) == list:
            return [self._get_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._get_data_on_s3(v)
        return data


def _convert_response(resp):
    if type(resp) is tuple:
        status_code = resp[1]
        if len(resp) == 2:
            return Response(body=resp[0], status_code=status_code)
        else:
            return Response(body=resp[0], status_code=status_code, headers=resp[2])

    elif not isinstance(resp, Response):
        return Response(body=resp)

    return resp


class Boto3Mixin:

    def __init__(self, service='s3', profile_name=None, env_var_access_key='aws_access_key_id',
                 env_var_secret_key='aws_secret_access_key', env_var_region='aws_region',
                 **kwargs):
        super().__init__(**kwargs)
        self.__session__ = self.__client__ = None
        self.__service = service
        self.__env_var_access_key = env_var_access_key
        self.__env_var_secret_key = env_var_secret_key
        self.__env_var_region = env_var_region

        self.__profile_name = profile_name

    @property
    def aws_access_key(self):
        value = os.getenv(self.__env_var_access_key)
        if not value:
            print(f"{self.__env_var_access_key} not defined in environment")
            raise EnvironmentError(f"{self.__env_var_access_key} not defined in environment")
        return value

    @property
    def aws_secret_access_key(self):
        value = os.getenv(self.__env_var_secret_key)
        if not value:
            print(f"{self.__env_var_secret_key} not defined in environment")
            raise EnvironmentError(f"{self.__env_var_secret_key} not defined in environment")
        return value

    @property
    def region_name(self):
        value = os.getenv(self.__env_var_region)
        if not value:
            print(f"{self.__env_var_region} not defined in environment")
            raise EnvironmentError(f"{self.__env_var_region} not defined in environment")
        return value

    @property
    def client(self):
        if self.__client__ is None:
            self.__client__ = self.__session.client(self.__service)
        return self.__client__

    @property
    def __session(self):
        if self.__session__ is None:
            region_name = self.region_name
            if self.__profile_name is not None:
                try:
                    self.__session__ = boto3.Session(profile_name=self.__profile_name, region_name=region_name)
                except Exception:
                    print(f"Cannot create session for profile {self.__profile_name} and region {region_name}")
                    raise
            else:
                access_key = self.aws_access_key
                secret_key = self.aws_secret_access_key
                try:
                    self.__session__ = boto3.Session(access_key, secret_key, region_name=region_name)
                except Exception:
                    print(f"Cannot create session for key {access_key}, secret {secret_key} and region {region_name}")
                    raise
        return self.__session__


class FileParam:

    def __init__(self, file, mime_type):
        self.file = file
        self.mime_type = mime_type

    def __repr__(self):
        if self.mime_type:
            return f'FileParam({self.file.name}, {self.mime_type})'
        return f'FileParam({self.file.name})'


class AwsS3Session(Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__(service='s3', **kwargs)


class AwsSFNSession(Boto3Mixin):

    def __init__(self, **kwargs):
        super().__init__(service='stepfunctions', **kwargs)



import os
import sqlalchemy
import re

from chalice import BadRequestError
from sqlalchemy import create_engine, text, MetaData, or_, and_, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from collections import defaultdict
from typing import List

from ..coworks import TechMicroService


class SqlMicroService(TechMicroService):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dialect = self.host = self.port = self.dbname = self.user = self.password = None
        self.engine = None
        self.session = None

        @self.before_first_request
        def check_env_vars():
            self.dialect = os.getenv('DIALECT')
            if not self.dialect:
                raise EnvironmentError('DIALECT not defined in environment')
            self.host = os.getenv('HOST')
            if not self.host:
                raise EnvironmentError('HOST not defined in environment')
            self.port = os.getenv('PORT')
            self.dbname = os.getenv('DB_NAME')
            if not self.dbname:
                raise EnvironmentError('DB_NAME not defined in environment')
            self.user = os.getenv('USER')
            if not self.user:
                raise EnvironmentError('USER not defined in environment')
            self.password = os.getenv('PASSWD')
            if not self.password:
                self.password = ''

        @self.before_first_request
        def engine():
            if self.port is None:
                if self.dialect == 'mysql':
                    self.port = 3306
                elif self.dialect == 'postgres':
                    self.port = 5432
            self.engine = create_engine(
                f'{self.dialect}://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}',
                echo=False)

    def get_version(self):
        """Returns SQLAlchemy version."""
        return sqlalchemy.__version__

    def get_fetch(self, query: str = None, **kwargs):
        conn = self.engine.connect()
        rows = conn.execute(text(query), **kwargs).fetchall()
        return [dict(row) for row in rows]

    def reflect_tables(self, schema, tables):
        metadata = MetaData(bind=self.engine)
        return list(map(lambda table: Table(table, metadata, autoload=True, autoload_with=self.engine, schema=schema), tables))



from unittest.mock import Mock, MagicMock

import pytest

from coworks import BizFactory, Every, At


def test_biz_reactor():
    fact = BizFactory('step_function')
    fact.create('test', Every(5, Every.MINUTES))
    assert len(fact.trigger_sources) == 1
    assert fact.trigger_sources[0]['name'] == 'step_function-test'
    assert fact.trigger_sources[0]['source'] == 'every'
    assert fact.trigger_sources[0]['value'] == 'rate(5 minutes)'
    with pytest.raises(Exception) as execinfo:
        fact.create('test', At(5, 10))
    assert len(fact.trigger_sources) == 1
    assert str(execinfo.value.args[0]) == 'BadRequestError: Biz microservice test already defined for step_function'
    fact.create('test2', At(5, 10, day_of_week='*'))
    assert len(fact.trigger_sources) == 2
    assert fact.trigger_sources[1]['name'] == 'step_function-test2'
    assert fact.trigger_sources[1]['source'] == 'at'
    assert fact.trigger_sources[1]['value'] == 'cron(5 10 None None * None)'


def test_biz_factory_undefined():
    fact = BizFactory('sfn_name')
    fact.__sfn_client__ = Mock()
    fact.sfn_client.list_state_machines = \
        MagicMock(return_value={'stateMachines': [
            {'name': 'sfn_name1', 'stateMachineArn': 'arn1'},
            {'name': 'sfn_name2', 'stateMachineArn': 'arn2'}
        ], 'nextToken': None})

    with pytest.raises(Exception) as execinfo:
        fact({'detail-type': 'Scheduled Event',
              'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name-every1']}, {})
    assert str(execinfo.value.args[0]) == 'BadRequestError: Undefined step function : sfn_name'

    fact = BizFactory('sfn_name')
    fact.__sfn_client__ = Mock()
    fact.sfn_client.list_state_machines = \
        MagicMock(return_value={'stateMachines': [
            {'name': 'sfn_name', 'stateMachineArn': 'arn1'},
            {'name': 'sfn_name2', 'stateMachineArn': 'arn2'}
        ], 'nextToken': None})
    with pytest.raises(Exception) as execinfo:
        fact({'detail-type': 'Scheduled Event',
              'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name-every1']}, {})
    assert str(execinfo.value.args[0]) == 'BadRequestError: Unregistered biz : every1'

    fact.create('every', Every(5, Every.MINUTES), data={"key1": "value1"})
    with pytest.raises(Exception) as execinfo:
        fact({'detail-type': 'Scheduled Event',
              'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name-every1']}, {})
    assert str(execinfo.value.args[0]) == 'BadRequestError: Unregistered biz : every1'


def test_biz_factory():
    fact = BizFactory('sfn_name1')
    fact.__sfn_client__ = Mock()
    fact.__sfn_arn__ = "arn1"

    fact.create('every1', Every(5, Every.MINUTES), data={"key1": "value1"})
    assert len(fact.biz) == 1
    assert len(fact.trigger_sources) == 1
    assert 'every1' in fact.biz
    fact.create('at2', At('0/10', '*', day_of_month='?', day_of_week='MON-FRI'), data={"key2": "value2"})
    assert len(fact.biz) == 2
    assert len(fact.trigger_sources) == 2
    assert 'at2' in fact.biz

    fact({'detail-type': 'Scheduled Event',
          'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name1-every1']}, {})
    fact.sfn_client.start_execution.assert_called_once_with(input='{"key1": "value1"}', stateMachineArn='arn1')
    fact({'detail-type': 'Scheduled Event',
          'resources': ['arn:aws:events:eu-west-1:123456789:rule/sfn_name1-at2']}, {})
    fact.sfn_client.start_execution.assert_called_with(input='{"key2": "value2"}', stateMachineArn='arn1')


def test_biz_data():
    fact = BizFactory('step_function')
    fact.__sfn_client__ = Mock()
    fact.__sfn_arn__ = "arn"

    fact.create('every1', Every(5, Every.MINUTES))
    fact({'detail-type': 'Scheduled Event',
          'resources': ['arn:aws:events:eu-west-1:123456789:rule/step_function-every1']}, {})
    fact.sfn_client.start_execution.assert_called_once_with(input='{}', stateMachineArn='arn')
    fact.create('every2', Every(10, Every.HOURS), data={'key': 'value'})
    fact({'detail-type': 'Scheduled Event',
          'resources': ['arn:aws:events:eu-west-1:123456789:rule/step_function-every2']}, {})
    fact.sfn_client.start_execution.assert_called_with(input='{"key": "value"}', stateMachineArn='arn')

    assert len(fact.trigger_sources) == 2
    assert [t['name'] for t in fact.trigger_sources] == ['step_function-every1', 'step_function-every2']
    assert [t['value'] for t in fact.trigger_sources] == ['rate(5 minutes)', 'rate(10 hours)']



import inspect

from aws_xray_sdk.core import xray_recorder


def class_auth_methods(obj):
    """Returns the auth method from the class if exists."""
    methods = inspect.getmembers(obj.__class__, lambda x: inspect.isfunction(x))

    for name, func in methods:
        if name == 'auth':
            return func
    return None


def class_rest_methods(obj):
    """Returns the list of methods from the class."""
    methods = inspect.getmembers(obj.__class__, lambda x: inspect.isfunction(x))

    res = []
    for name, func in methods:
        if name == 'get' or name.startswith('get_'):
            res.append(('get', func))
        elif name == 'post' or name.startswith('post_'):
            res.append(('post', func))
        elif name == 'put' or name.startswith('put_'):
            res.append(('put', func))
        elif name == 'delete' or name.startswith('delete_'):
            res.append(('delete', func))
        elif name == 'patch' or name.startswith('patch_'):
            res.append(('patch', func))
        elif name == 'options' or name.startswith('options_'):
            res.append(('options', func))
    return res


def class_attribute(obj, name: str = None, defaut=None):
    """Returns the list of attributes from the class or the attribute if name parameter is defined
    or default value if not found."""
    attributes = inspect.getmembers(obj.__class__, lambda x: not inspect.isroutine(x))

    if not name:
        return attributes

    filtered = [a[1] for a in attributes if a[0] == name]
    return filtered[0] if filtered else defaut


def trim_underscores(name):
    while name.startswith('_'):
        name = name[1:]
    while name.endswith('_'):
        name = name[:-1]
    return name


def begin_xray_subsegment(subsegment_name):
    if xray_recorder.in_segment().segment is not None:
        return xray_recorder.begin_subsegment(subsegment_name)
    return None


def end_xray_subsegment():
    if xray_recorder.in_subsegment().subsegment is not None:
        return xray_recorder.end_subsegment()



import pytest
import os

from coworks.pytest.fixture import local_server_factory as factory
from tests.mockup import email_mock, smtp_mock, boto3_mock


@pytest.fixture()
def email_mock_fixture():
    yield email_mock


@pytest.fixture()
def smtp_mock_fixture():
    yield smtp_mock


@pytest.fixture()
def boto3_mock_fixture():
    yield boto3_mock


local_server_factory = factory


def pytest_sessionstart():
    example_dir = os.getenv('EXAMPLE_DIR')
    if not os.path.exists(example_dir):
        raise pytest.UsageError("Undefined example folder: {example_dir}.")




import inspect
import pathlib
from abc import ABC, abstractmethod
import sys

from jinja2 import Environment, PackageLoader, select_autoescape, TemplateNotFound


class WriterError(Exception):
    ...


class Writer(ABC):

    def __init__(self, app=None, name=None):
        self.output = sys.stdout
        self.error = sys.stderr
        self.name = name

        #: A list of functions that will be called before or after the export command.
        self.before_funcs = []
        self.after_funcs = []

        if app is not None:
            self.app = app
            self.init_app(app)

    def init_app(self, app):
        if 'writers' not in app.extensions:
            app.extensions['writers'] = {}
        app.extensions['writers'][self.name] = self

    def export(self, output=None, error=None, **kwargs):
        """ Called when the export command of the Coworks CLI is called.
        :param output: output stream.
        :param error: error stream.
        :param kwargs: environment parameters for export.
        :return: None
        """
        if output is not None:
            self.output = open(output, 'w+') if type(output) is str else output
        if error is not None:
            self.error = open(error, 'w+') if type(error) is str else error

        for func in self.before_funcs:
            func()

        self._export_header(**kwargs)
        self._export_content(**kwargs)

        for func in self.after_funcs:
            func()

    def before_export(self, f):
        """Registers a function to be run before the export command.
        :param f: function called before the export commaand
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.after_funcs.append(f)
        return f

    def after_export(self, f):
        """Registers a function to be run after the export command.
        :param f: function called after the export commaand
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.after_funcs.append(f)
        return f

    def _export_header(self, **kwargs):
        ...

    @abstractmethod
    def _export_content(self, **kwargs):
        """ Main export function.
        :param kwargs: Environment parameters for export.
        :return: None.

        Abstract method which must be redefined in any subclass. The content should be written in self.output.
        """

    def _format(self, content):
        return content


class TemplateWriter(Writer):

    def __init__(self, app=None, name=None, data=None, template_filenames=None, env=None):
        super().__init__(app, name)
        self.data = data or {}
        self.template_filenames = template_filenames or self.default_template_filenames
        self.env = env or Environment(
            loader=PackageLoader("coworks.cws.writer"),
            autoescape=select_autoescape(['html', 'xml'])
        )

    @property
    @abstractmethod
    def default_template_filenames(self):
        ...

    def _export_content(self, module_name='app', handler_name='app', project_dir='.', **kwargs):
        module_path = module_name.split('.')
        data = {
            'writer': self,
            'project_dir': project_dir,
            'module': module_name,
            'module_path': pathlib.PurePath(*module_path),
            'module_dir': pathlib.PurePath(*module_path[:-1]),
            'module_file': module_path[-1],
            'handler': handler_name,
            'app': self.app,
            'app_name': self.app.app_name,
        }
        data.update(self.data)
        try:
            for template_filename in self.template_filenames:
                template = self.env.get_template(template_filename)
                print(self._format(template.render(**data)), file=self.output)
        except TemplateNotFound as e:
            raise WriterError(f"Cannot find template {str(e)}")
        except Exception as e:
            raise WriterError(e)


class ListWriter(TemplateWriter):

    def __init__(self, app=None, name='list', **kwargs):
        super().__init__(app, name, **kwargs)

    @property
    def default_template_filenames(self):
        return ['list.j2']

    def _export_header(self, **kwargs):
        pass


class OpenApiWriter(TemplateWriter):
    """Export the microservice in swagger format."""

    def __init__(self, app=None, name='openapi', **kwargs):
        super().__init__(app, name, **kwargs)

    @property
    def default_template_filenames(self):
        return ['openapi.yml']

    def _export_header(self, **kwargs):
        print("# Do NOT edit this file as it is auto-generated by cws\n", file=self.output)

    @property
    def pathes(self):
        pathes = {}
        for route, entry in self.app.routes.items():
            entries = {}
            for method, infos in entry.items():
                func = self.app.entries[route]
                fullargspec = inspect.getfullargspec(func[1])
                args = fullargspec.args[1:]
                defaults = fullargspec.defaults or []
                len_defaults = len(defaults)
                entry = {
                    'description': func.__doc__,
                    'parameters': []
                }
                for index, arg in enumerate(args[1:-len_defaults]):
                    entry['parameters'].append({
                        'name': f"_{index}:{arg}",
                        'in': 'path',
                    })
                for arg in args[-len_defaults:]:
                    entry['parameters'].append({
                        'name': arg,
                        'in': 'query',
                    })
                for varkw in fullargspec.varkw or []:
                    entry['parameters'].append({
                        'name': varkw,
                        'in': 'query',
                    })
                entries[method.lower()] = entry
            pathes[route] = entries
        return pathes


def reduce_not_none(data):
    """Remove all entries where value is None."""
    return dict((k, v) for k, v in data.items() if v is not None)



from coworks import Blueprint


class BP(Blueprint):

    def get_test(self, index):
        return f"blueprint test {index}"

    def get_extended_test(self, index):
        return f"blueprint extended test {index}"



import json
import os
import pathlib
from abc import ABC, abstractmethod

import yaml

from .writer import Writer, WriterError

INITIAL_STATE_NAME = "Init"
LAMBDA_ERROR_FALLBACK = "MicroServiceErrorFallback"


class StepFunctionWriter(Writer):

    def __init__(self, app=None, name='sfn', extension='yml'):
        super().__init__(app, name)
        self.extension = extension

    def _export_content(self, module_name='app', handler_name='biz', project_dir='.', **kwargs):
        module_path = module_name.split('.')
        step_functions = {}
        errors = {}
        sfn_name = self.app.sfn_name
        filename = pathlib.Path(project_dir, *module_path[:-1]) / f"{sfn_name}.{self.extension}"
        try:
            sfn = StepFunction(sfn_name, filename)
            step_functions[sfn_name] = sfn.generate()
        except WriterError as e:
            errors[sfn_name] = str(e)

        if errors:
            for sfn_name, error in errors.items():
                print(f"Error in {sfn_name}: {error}", file=self.error)
            raise WriterError()
        else:
            for idx, (sfn_name, sfn) in enumerate(step_functions.items()):
                if idx > 0:
                    print("---", file=self.output)
                print(json.dumps(sfn, indent=2), file=self.output)


class StepFunction:

    def __init__(self, sfn_name, filepath):
        self.name = sfn_name
        self.all_states = []
        try:
            with filepath.open() as file:
                self.data = yaml.load(file, Loader=yaml.SafeLoader)
                if not self.data:
                    raise WriterError(f"The content of the {sfn_name} microservice seems to be empty.")
        except FileNotFoundError:
            path = pathlib.Path(filepath)
            if not path.is_absolute():
                path = pathlib.Path(f"{os.getcwd()}/{filepath}")
            raise WriterError(f"The source for the {sfn_name} microservice should be found in {path}.")
        except Exception as e:
            raise WriterError(str(e))

        self.global_catch = 'catch' in self.data

    def generate(self):
        first = PassState(self, {'name': INITIAL_STATE_NAME})

        self.all_states.append(first)
        states = State.get_or_raise(self.data, 'states')
        self.add_actions(self.all_states, states)

        first.add_next_or_end('Next', self.all_states[1].name)

        # global catch state (no automatic catch in corresponding ations)
        if self.global_catch:
            error_fallback = PassState(self, {'name': LAMBDA_ERROR_FALLBACK})
            self.all_states.append(error_fallback)
            self.add_actions(self.all_states, self.data['catch'], no_catch=True)

        return {
            "Version": "1.0",
            "Comment": self.data.get('comment', self.name),
            "StartAt": "Init",
            "States": {state.name: state.state for state in self.all_states}
        }

    def add_actions(self, states, actions, no_catch=False):
        if not actions:
            raise WriterError("Actions list cannot be empty.")

        previous_state = states[-1] if len(states) > 0 else None
        for action in actions:
            state = self.new_state(action, previous_state, no_catch=no_catch)
            states.append(state)
            previous_state = state

            if action == actions[-1] and 'Next' not in state.state:
                state.add_next_or_end('End', True)

            states.extend(state.catch_states)

    def new_state(self, action, previous_state, no_catch=False):
        def add_next_to_previous_state():
            if isinstance(previous_state, ChoiceState):
                if 'Default' not in previous_state.state:
                    previous_state.state['Default'] = state.name
            else:
                if 'Next' not in previous_state.state and 'End' not in previous_state.state:
                    previous_state.state['Next'] = state.name

        if 'choices' in action:
            state = ChoiceState(self, action)
        elif 'fail' in action:
            state = FailState(self, action)
        elif 'pass' in action:
            state = PassState(self, action)
        elif 'success' in action:
            state = SuccessState(self, action)
        elif 'tech' in action:
            state = TechState(self, action, no_catch=no_catch)
        elif 'wait' in action:
            state = WaitState(self, action)
        else:
            raise WriterError(f"Undefined type of action for {action}")

        if previous_state is not None:
            add_next_to_previous_state()

        return state

    def __repr__(self):
        return f"{self.name} : {self.all_states}"


class State(ABC):

    def __init__(self, sfn, action, **kwargs):
        self.sfn = sfn
        self.action = action or {}
        self.state = {**kwargs}
        self.catch_states = []

    @property
    def name(self):
        return self.action.get('name', f'The action {self.action} has no name')

    @property
    def slug(self):
        return "_".join(self.name.split()).lower()

    @staticmethod
    def get_or_raise(action, key):
        """Get the value at key or raise an error if the key doesn't exist."""
        try:
            return action[key]
        except (KeyError, TypeError):
            raise WriterError(f"The key {key} is missing for {action}")

    @staticmethod
    def get_goto(action_or_choice):
        goto = action_or_choice.get('goto')
        if goto == "End":
            return "End", True
        elif goto is not None:
            return "Next", goto
        return None

    def add_goto(self, action):
        entry = self.get_goto(action)
        if entry:
            key, value = entry
            self.add_next_or_end(key, value)

    def __repr__(self):
        return f"{self.name} : {json.dumps(self.state, indent=2)}"

    @abstractmethod
    def add_next_or_end(self, key, value):
        """Do nothing by default."""


class SuccessState(State):
    def __init__(self, sfn, action=None, **kwargs):
        kwargs.update({
            "Type": "Succeed",
        })
        super().__init__(sfn, action, **kwargs)

    def add_next_or_end(self, key, value):
        super().add_next_or_end(key, value)


class FailState(State):
    def __init__(self, sfn, action, **kwargs):
        super().__init__(sfn, action, Type="Fail", **kwargs)
        self.state['Cause'] = action.get('cause', "Undefined reason")
        self.state['Error'] = action.get('error', "Undefined error")

    def add_next_or_end(self, key, value):
        super().add_next_or_end(key, value)


class ChoiceState(State):

    def __init__(self, sfn, action, **kwargs):
        super().__init__(sfn, action, Type="Choice", **kwargs)

        choices = self.get_or_raise(action, 'choices')
        if not choices:
            raise WriterError(f"The list of choices may not be empty {action}")

        self.state['Choices'] = self.create_choices_sequence(choices)

        if 'default' in action:
            self.state['Default'] = action.get('default')

    def create_choices_sequence(self, choices):

        def create_condition(data):
            var = self.get_or_raise(data, 'var')
            oper = self.get_or_raise(data, 'oper')
            val = self.get_or_raise(data, 'value')
            return {
                'Variable': var,
                oper: val,
            }

        switch = []
        for choice in choices:
            entry = self.get_goto(choice)
            if entry is None:
                raise KeyError(f"The key goto is missing for {choice}")

            key, value = entry
            if 'not' in choice:
                switch.append({"Not": create_condition(choice['not']), key: value})
            elif 'or' in choice:
                switch.append({"Or": create_condition(choice['or']), key: value})
            elif 'and' in choice:
                switch.append({"And": create_condition(choice['and']), key: value})
            else:
                switch.append({**create_condition(choice), key: value})
        return switch

    def add_next_or_end(self, key, value):
        super().add_next_or_end(key, value)


class PassState(State):
    def __init__(self, sfn, action, **kwargs):
        kwargs.setdefault("Type", "Pass")
        super().__init__(sfn, action, **kwargs)

    def add_next_or_end(self, key, value):
        self.state[key] = value


class EndState(PassState):
    def __init__(self, sfn, action, **kwargs):
        super().__init__(sfn, action, **kwargs)
        self.state['End'] = True

    def add_next_or_end(self, key, value):
        super().add_next_or_end(key, value)


class WaitState(PassState):
    def __init__(self, sfn, action, **kwargs):
        super().__init__(sfn, action, Type="Wait", **kwargs)
        self.state['Seconds'] = action['wait']
        self.add_goto(action)


class TechState(PassState):
    def __init__(self, sfn, action, **kwargs):
        self.no_catch = kwargs.pop('no_catch', False)
        super().__init__(sfn, action, Type="Task", **kwargs)

        tech_data = self.get_or_raise(action, 'tech')
        if not tech_data:
            raise WriterError(f"The content of tech action is empty")

        try:
            res = self.get_or_raise(tech_data, 'service')
            self.state['Resource'] = f"arn:aws:lambda:eu-west-1:935392763270:function:{res}"
            self.state["InputPath"] = f"$"
            self.state["ResultPath"] = f"$.{self.slug}.result"
            self.state["OutputPath"] = "$"
            self.state["Parameters"] = self.get_call_data(action, tech_data)

            if 'catch' not in action:
                if self.sfn.global_catch and not self.no_catch:
                    self.state["Catch"] = [{
                        "ErrorEquals": ["BadRequestError"],
                        "Next": LAMBDA_ERROR_FALLBACK
                    }]
            else:
                self.sfn.add_actions(self.catch_states, action['catch'])
                self.state["Catch"] = [{
                    "ErrorEquals": ["BadRequestError"],
                    "Next": self.catch_states[0].name
                }]
                if self.sfn.global_catch and "States.TaskFailed" not in [c["ErrorEquals"] for c in self.state["Catch"]]:
                    self.state["Catch"].append({
                        "ErrorEquals": ["States.TaskFailed"],
                        "Next": LAMBDA_ERROR_FALLBACK
                    })

        except KeyError as e:
            raise WriterError(f"The key {e} is missing for {action}")

        self.add_goto(action)

        # complementary informations
        timeout = tech_data.get('timeout')
        if timeout is not None:
            self.state["TimeoutSeconds"] = timeout

    @staticmethod
    def get_call_data(action, tech_data):

        # get route and method
        route = method = None
        has_get = 'get' in tech_data
        has_post = 'post' in tech_data
        has_put = 'put' in tech_data
        if has_get + has_post + has_put > 1:
            raise WriterError(f"Too many methods defined for {action}")
        if has_get:
            route = tech_data['get']
            method = 'GET'
        if has_post:
            route = tech_data['post']
            method = 'POST'
        if has_put:
            route = tech_data['put']
            method = 'PUT'
        if route is None:
            raise WriterError(f"No route defined for {action}")

        # get call parameters
        uri_params = tech_data.get('uri_params')
        query_params = tech_data.get('query_params')
        body = tech_data.get('body')
        form_data = tech_data.get('form-data')
        content_type = "application/json" if form_data is None else "multipart/form-data"

        return {
            "type": "CWS_SFN",
            "resource": route,
            "path": route,
            "requestContext": {
                "resourcePath": route,
                'httpMethod': method,
            },
            "multiValueQueryStringParameters": query_params,
            "pathParameters": uri_params,
            "headers": {
                "Accept-Encoding": "gzip,deflate",
                "Content-Type": content_type
            },
            "body": body if body else None,
            "form-data": TechState.validate_form_data(form_data) if form_data else None,
            "stageVariables": None,
        }

    @staticmethod
    def validate_form_data(form_data):
        return form_data



import logging
import os
from http.client import BadStatusLine
from xmlrpc import client

from aws_xray_sdk.core import xray_recorder
from chalice import ChaliceViewError
from chalice import NotFoundError, BadRequestError
from pyexpat import ExpatError

from .. import Blueprint
from ..coworks import TechMicroService


class OdooMicroService(TechMicroService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.url = self.db = self.username = self.password = self.models_url = self.api_uid = self.logger = None

    def get_model(self, model: str, searched_field: str, searched_value=None, fields=None, searched_type="str",
                  ensure_one=False, **kwargs):
        """Returns the list of objects or the object which searched_field is equal to the searched_value."""
        if not searched_value:
            return BadRequestError(f"{searched_field} not defined.")

        convert = str
        if searched_type == "int":
            convert = int
        results = self.search(model, [[(searched_field, '=', convert(searched_value))]],
                              fields=fields if fields else [],
                              **kwargs)

        if ensure_one:
            return self._ensure_one(results)

        if not results:
            raise NotFoundError("No object found")
        return results

    def put_model(self, model, data=None):
        return self.create(model, data)

    def put_model_(self, model, _id, data=None):
        """Replaces the object of the model referenced by this id."""
        return self.write(model, _id, data)

    def delete_model(self, model, _id, dry=False):
        """Deletes the object of the model referenced by this id."""
        return self.execute_kw(model, 'unlink', [[int(_id)]], dry=dry)

    def get_field(self, model, searched_field, searched_value, returned_field='id'):
        """Returns the value of the object which searched_field is equal to the searched_value."""
        value = self.get_model(model, searched_field, searched_value, fields=[returned_field], ensure_one=True)
        return value[returned_field]

    def get_id(self, model, searched_field, searched_value):
        """Returns the id of the object which searched_field is equal to the searched_value."""
        return self.get_field(model, searched_field, searched_value)

    def post_call(self, model: str, method: str, params_or_filters=None, options=None):
        params_or_filters = params_or_filters or [[]]
        options = options or {}
        return self.execute_kw(model, method, [params_or_filters], options)

    def execute_kw(self, model: str, method: str, *args, dry=False):
        try:
            if not model:
                raise ChaliceViewError("Model undefined")

            if not self.api_uid:
                self._connect()
            self.logger.info(f'Execute_kw : {model}, {method}, {list(args)}')
            if dry:
                return

            try:
                subsegment = xray_recorder.begin_subsegment(f"Quering ODOO")
                with client.ServerProxy(self.models_url, allow_none=True) as models:
                    if subsegment:
                        subsegment.put_metadata('model', model)
                        subsegment.put_metadata('method', method)
                        for index, arg in enumerate(args):
                            subsegment.put_metadata(f'arg{index}', arg)
                    return models.execute_kw(self.db, self.api_uid, self.password, model, method, *args)
            finally:
                xray_recorder.end_subsegment()
        except (BadStatusLine, ExpatError):
            self.logger.debug(f'Retry execute_kw : {model} {method} {args}')
            with client.ServerProxy(self.models_url) as models:
                return models.execute_kw(self.db, self.api_uid, self.password, model, method, *args)
        except Exception as e:
            raise ChaliceViewError(str(e))

    def search(self, model, filters: list, fields=None, offset=None, limit=None, order=None) -> list:
        options = {}
        if fields:
            options["fields"] = fields if type(fields) is list else [fields]
        options.setdefault("offset", int(offset) if offset else 0)
        options.setdefault("limit", int(limit) if limit else 50)
        options.setdefault("order", order if order else 'id asc')
        return self.execute_kw(model, 'search_read', filters, options)

    def create(self, model, data: dict, dry=False):
        return self.execute_kw(model, 'create', [self._replace_tuple(data)], dry=dry)

    def write(self, model, _id, data: dict, dry=False):
        return self.execute_kw(model, 'write', [[_id], self._replace_tuple(data)], dry=dry)

    def _connect(self, url=None, database=None, username=None, password=None):

        # initialize connection informations
        self.url = url or os.getenv('ODOO_URL')
        if not self.url:
            raise EnvironmentError('ODOO_URL must be set before anything else!')
        self.db = database or os.getenv('ODOO_DB')
        if not self.db:
            raise EnvironmentError('ODOO_DB must be set before anything else!')
        self.username = username or os.getenv('ODOO_USERNAME')
        if not self.username:
            raise EnvironmentError('ODOO_USERNAME must be set before anything else!')
        self.password = password or os.getenv('ODOO_PASSWORD')
        if not self.password:
            raise EnvironmentError('ODOO_PASSWORD must be set before anything else!')

        self.logger = logging.getLogger('odoo')

        try:
            xray_recorder.begin_subsegment(f"Connecting ODOO")

            # initialize xml connection to odoo
            common = client.ServerProxy(f'{self.url}/xmlrpc/2/common')
            self.api_uid = common.authenticate(self.db, self.username, self.password, {})
            if not self.api_uid:
                raise Exception(f'Odoo connection parameters are wrong')
            self.models_url = f'{self.url}/xmlrpc/2/object'
        except Exception:
            raise Exception(f'Odoo interface variables wrongly defined.')
        finally:
            xray_recorder.end_subsegment()

    @staticmethod
    def _ensure_one(results) -> dict:
        """Ensure only only one in the result list and returns it."""
        if len(results) == 0:
            raise NotFoundError(f"No object found.")
        if len(results) > 1:
            raise NotFoundError(
                f"More than one object ({len(results)}) founds : ids={[o.get('id') for o in results]}")
        return results[0]

    def _replace_tuple(self, struct: dict) -> dict:
        """For data from JSON, tuple are defined with key surronded by paranthesis."""
        for k, value in struct.items():
            if isinstance(value, dict):
                self._replace_tuple(value)
            else:
                if k.startswith('(') and type(value) is list:
                    del struct[k]
                    struct[k[1:-1]] = [tuple(v) for v in value]
        return struct


class OdooBlueprint(Blueprint):
    def __init__(self, model, common_filters=None, **kwargs):
        super().__init__(**kwargs)
        self._model = model
        self._common_filters = common_filters if common_filters else []

    def search(self, filters: list, fields=None, offset=None, limit=None, order=None, **options):
        filters = [self._common_filters + f for f in filters]
        return self.current_app.search(self._model, filters, fields, offset, limit, order, **options)

    def create(self, data, dry=False):
        return self.current_app.put_model(self._model, data, dry=dry)

    def write(self, _id, data, dry=False):
        return self.current_app.put_model(self._model, _id, data, dry=dry)

    def delete(self, _id, dry=False):
        return self.current_app.delete_model(self._model, _id, dry=dry)


class UserBlueprint(OdooBlueprint):

    def __init__(self, import_name='user', **kwargs):
        super().__init__("res.users", import_name=import_name, **kwargs)


class PartnerBlueprint(OdooBlueprint):

    def __init__(self, import_name='partner', **kwargs):
        super().__init__("res.partner", import_name=import_name, **kwargs)


class CustomerBlueprint(PartnerBlueprint):

    def __init__(self, import_name='customer', **kwargs):
        super().__init__(common_filters=[('customer', '=', True)], import_name=import_name, **kwargs)


class SupplierBlueprint(PartnerBlueprint):

    def __init__(self, import_name='supplier', **kwargs):
        super().__init__(common_filters=[('supplier', '=', True)], import_name=import_name, **kwargs)


class ProductBlueprint(OdooBlueprint):

    def __init__(self, import_name='product', **kwargs):
        super().__init__("product.product", import_name=import_name, **kwargs)



import contextlib
import socket
from threading import Thread, Event

from chalice.config import Config
from chalice.local import LocalDevServer


class ThreadedLocalServer(Thread):
    threaded_servers = {}

    def __init__(self, *, port=None, host='localhost'):
        super(ThreadedLocalServer, self).__init__()
        self._app_object = None
        self._config = None
        self._host = host
        self._port = port or self.unused_tcp_port()
        self._server = None
        self._server_ready = Event()

    def wait_for_server_ready(self):
        self._server_ready.wait()

    def configure(self, app_object, config=None):
        self._app_object = app_object
        self._config = config if config else Config()

    def run(self):
        self._server = LocalDevServer(self._app_object, self._config, self._host, self._port)
        self._server_ready.set()
        self._server.serve_forever()

    def make_call(self, method, path, timeout=0.5, **kwarg):
        self._server_ready.wait()
        return method('http://{host}:{port}{path}'.format(
            path=path, host=self._host, port=self._port), timeout=timeout, **kwarg)

    def shutdown(self):
        if self._server is not None:
            self._server.server.shutdown()

    @classmethod
    def unused_tcp_port(cls):
        with contextlib.closing(socket.socket()) as sock:
            sock.bind(('localhost', 0))
            return sock.getsockname()[1]



from .csv import CSVMicroService
from .mail import MailMicroService
from .odoo import OdooMicroService
from .s3 import S3MicroService
from .jinja import JinjaRenderMicroService
from .sql import SqlMicroService



import inspect
import json
import logging
import sys
import traceback
from functools import update_wrapper
from threading import Lock
from typing import Dict

from aws_xray_sdk.core import xray_recorder
from chalice import AuthResponse, BadRequestError, Rate, Cron
from chalice import Chalice, Blueprint as ChaliceBlueprint

from requests_toolbelt.multipart import MultipartEncoder

from .config import Config
from .mixins import CoworksMixin, AwsSFNSession
from .utils import begin_xray_subsegment, end_xray_subsegment
from .utils import class_auth_methods, class_rest_methods, class_attribute, trim_underscores

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class TechMicroService(CoworksMixin, Chalice):
    """Simple tech microservice created directly from class."""

    def __init__(self, app_name: str = None, config: Config = None, **kwargs):
        """ Initialize a technical microservice.
        :param app_name: Name used to identify the resource.
        :param config: Deployment configuration.
        :param kwargs: Other Chalice parameters.
        """
        app_name = app_name or self.__class__.__name__
        authorizer = kwargs.pop('authorizer', None)

        super().__init__(app_name, **kwargs)
        self.experimental_feature_flags.update([
            'BLUEPRINTS'
        ])

        self.config = config or Config()

        self.__auth__ = self._create_auth_proxy(self, authorizer) if authorizer else None

        # Blueprints added by names.
        self.blueprints = {}

        # Extensions defined by type's name (xriters, ..)
        self.extensions = {}

        # A list of functions that will be called at the first activation.
        # To register a function, use the :meth:`before_first_request` decorator.
        self.before_first_activation_funcs = []

        # A list of functions that will be called at the beginning of each activation.
        # To register a function, use the :meth:`before_activation` decorator.
        self.before_activation_funcs = []

        self.after_activation_funcs = []

        self._got_first_activation = False
        self._before_activation_lock = Lock()

        self.entries = {}
        self._add_route()
        self.sfn_call = False

        if "pytest" in sys.modules:
            xray_recorder.configure(context_missing="LOG_ERROR")

    @property
    def component_name(self):
        return class_attribute(self, 'url_prefix', '')

    @property
    def ms_type(self):
        return 'tech'

    @property
    def lambda_zip_file(self):
        return None

    @property
    def layer_zip_file(self):
        return None

    def register_blueprint(self, blueprint, **kwargs):
        if 'name_prefix' not in kwargs:
            kwargs['name_prefix'] = blueprint.component_name
        if 'url_prefix' not in kwargs:
            kwargs['url_prefix'] = f"/{blueprint.component_name}"

        self._add_route(blueprint, url_prefix=kwargs['url_prefix'])
        super().register_blueprint(blueprint, **kwargs)
        self.blueprints[blueprint.import_name] = blueprint

    def run(self, host='127.0.0.1', port=8000, project_dir='.', stage=None, debug=True):
        # chalice.cli and .cws packages not defined in deployment
        from chalice.cli import DEFAULT_STAGE_NAME
        from .cws.factory import CwsCLIFactory

        if debug:
            logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(message)s')

        stage = stage or DEFAULT_STAGE_NAME
        factory = CwsCLIFactory(self, project_dir, debug=debug)
        config = factory.mock_config_obj(self, stage)
        factory.run_local_server(self, config, host, port)

    def _add_route(self, component=None, url_prefix=None):
        if component is None:
            component = self

            # Adds class authorizer for every entries (if not already added before)
            auth = class_auth_methods(component)
            if auth and component.__auth__ is None:
                auth = TechMicroService._create_auth_proxy(component, auth)
                component.__auth__ = auth
        else:
            # Adds the current_app auth for blueprints
            auth = self.__auth__

        # Adds entrypoints
        methods = class_rest_methods(component)
        for method, func in methods:
            if func.__name__ == method:
                route = f"{component.component_name}"
            else:
                name = func.__name__[len(method) + 1:]
                name = trim_underscores(name)  # to allow several functions with same route but different args
                name = name.replace('_', '/')
                route = f"{component.component_name}/{name}" if component.component_name else f"{name}"
            args = inspect.getfullargspec(func).args[1:]
            defaults = inspect.getfullargspec(func).defaults
            varkw = inspect.getfullargspec(func).varkw
            if defaults:
                len_defaults = len(defaults)
                for index, arg in enumerate(args[:-len_defaults]):
                    route = route + f"/{{_{index}}}" if route else f"{{_{index}}}"
                kwarg_keys = args[-len_defaults:]
            else:
                for index, arg in enumerate(args):
                    route = route + f"/{{_{index}}}" if route else f"{{_{index}}}"
                kwarg_keys = {}

            proxy = component._create_rest_proxy(func, kwarg_keys, args, varkw)

            # complete all entries
            component.route(f"/{route}", methods=[method.upper()], authorizer=auth, cors=self.config.cors,
                            content_types=['multipart/form-data', 'application/json'])(proxy)
            if url_prefix:
                self.entries[f"{url_prefix}/{route}"] = (method.upper(), func)
            else:
                self.entries[f"/{route}"] = (method.upper(), func)

    @staticmethod
    def _create_auth_proxy(component, auth_method):

        def proxy(auth_activation):
            subsegment = begin_xray_subsegment(f"auth microservice")
            try:
                auth = auth_method(component, auth_activation)
                if subsegment:
                    subsegment.put_metadata('result', auth)
            except Exception as e:
                logger.info(f"Exception : {str(e)}")
                traceback.print_exc()
                if subsegment:
                    subsegment.add_exception(e, traceback.extract_stack())
                raise BadRequestError(str(e))
            finally:
                end_xray_subsegment()

            if type(auth) is bool:
                if auth:
                    return AuthResponse(routes=['*'], principal_id='user')
                return AuthResponse(routes=[], principal_id='user')
            elif type(auth) is list:
                return AuthResponse(routes=auth, principal_id='user')
            return auth

        proxy = update_wrapper(proxy, auth_method)
        proxy.__name__ = 'app'
        return component.authorizer(name='auth')(proxy)

    def __call__(self, event, context):
        with self._before_activation_lock:
            if not self._got_first_activation:
                for func in self.before_first_activation_funcs:
                    func()
                self._got_first_activation = True

        for func in self.before_activation_funcs:
            func()

        res = self.handler(event, context)

        for func in self.after_activation_funcs:
            func()

        return res

    def handler(self, event, context):

        # authorization call
        if event.get('type') == 'TOKEN':
            if self.debug:
                print(f"Calling {self.app_name} for authorization")

            if self.__auth__:
                return self.__auth__(event, context)

            print(f"Undefined authorization method for {self.app_name} ")
            raise Exception('Unauthorized')

        # step function call
        if event.get('type') == 'CWS_SFN':
            if self.debug:
                print(f"Calling {self.app_name} by step function")

            self.sfn_call = True
            content_type = event['headers']['Content-Type']
            if content_type == 'application/json':
                body = event.get('body')
                event['body'] = json.dumps(self._get_data_on_s3(body)) if body else body
            elif content_type == 'multipart/form-data':
                if event.get('form-data'):
                    multi_parts = MultipartEncoder(self._set_multipart_content(event.get('form-data')))
                    event['headers']['Content-Type'] = multi_parts.content_type
                    event['body'] = multi_parts.to_string()
            else:
                raise BadRequestError(f"Undefined content type {content_type} for Step Function call")
        else:
            if self.debug:
                print(f"Calling {self.app_name} with event {event}")

        res = super().__call__(event, context)

        if self.sfn_call:
            if res['statusCode'] < 200 or res['statusCode'] >= 300:
                raise BadRequestError(f"Status code is {res['statusCode']} : {res['body']}")
            try:
                res['body'] = self._set_data_on_s3(json.loads(res['body']))
            except json.JSONDecodeError:
                pass

        if self.debug:
            print(f"Call {self.app_name} returns {res}")

        return res

    def before_first_activation(self, f):
        """Registers a function to be run before the first activation of the microservice.

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.before_first_activation_funcs.append(f)
        return f

    def before_activation(self, f):
        """Registers a function to run before each activation of the microservice.
        :param f:  Function added to the list.
        :return: None.

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.before_activation_funcs.append(f)
        return f

    def after_activation(self, f):
        """Registers a function to be run after each activation of the microservice.

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.after_activation_funcs.append(f)
        return f


class BizFactory(TechMicroService):
    """Tech microservice to create, update and trigger biz microservices."""

    def __init__(self, sfn_name, **kwargs):
        super().__init__(app_name=sfn_name, **kwargs)

        self.aws_profile = self.__sfn_client__ = self.__sfn_arn__ = None
        self.sfn_name = sfn_name
        self.biz: Dict[str, BizMicroService] = {}

        @self.before_first_activation
        def check_sfn():
            return self.sfn_arn

    @property
    def ms_type(self):
        return 'biz'

    @property
    def trigger_sources(self):

        def to_dict(name, trigger):
            return {'name': f"{self.sfn_name}-{name}", **trigger.to_dict()}

        return [to_dict(name, biz.trigger) for name, biz in self.biz.items()]

    @property
    def sfn_client(self):
        if self.__sfn_client__ is None:
            session = AwsSFNSession(profile_name=self.aws_profile, env_var_access_key="AWS_RUN_ACCESS_KEY_ID",
                                    env_var_secret_key="AWS_RUN_SECRET_KEY", env_var_region="AWS_RUN_REGION")
            self.__sfn_client__ = session.client
        return self.__sfn_client__

    @property
    def sfn_arn(self):
        if self.__sfn_arn__ is None:
            res = self.sfn_client.list_state_machines()
            while True:
                for sfn in res['stateMachines']:
                    if sfn['name'] == self.sfn_name:
                        self.__sfn_arn__ = sfn['stateMachineArn']
                        return self.__sfn_arn__

                next_token = res.get('nextToken')
                if next_token is None:
                    raise BadRequestError(f"Undefined step function : {self.sfn_name}")

                res = self.sfn_client.list_state_machines(nextToken=next_token)
        return self.__sfn_arn__

    def get_sfn_name(self):
        """Returns the name of the associated step function."""
        return self.sfn_name

    def get_sfn_arn(self):
        """Returns the arn of the associated step function."""
        return self.sfn_arn

    def get_biz_names(self):
        """Returns the list of biz microservices defined in the factory."""
        return [name for name in self.biz]

    def post_trigger(self, biz_name, data=None):
        """Manual triggering a biz microservice."""
        if self.debug:
            print(f"Manual triggering: {biz_name}")

        data = data or {}
        try:
            return self.biz[biz_name](data, {})
        except KeyError:
            if self.debug:
                print(f"Cannot found {biz_name} in services {[k for k in self.biz.keys()]}")
            raise BadRequestError(f"Cannot found {biz_name} biz microservice")

    def create(self, biz_name, trigger=None, data: dict = None, **kwargs):
        """Creates a biz microservice. If the trigger is not defined the microservice can only be triggered manually."""

        if biz_name in self.biz:
            raise BadRequestError(f"Biz microservice {biz_name} already defined for {self.sfn_name}")

        self.biz[biz_name] = BizMicroService(self, data, trigger, app_name=biz_name, **kwargs)
        return self.biz[biz_name]

    def invoke(self, data):
        res = self.sfn_client.start_execution(stateMachineArn=self.sfn_arn, input=json.dumps(data if data else {}))
        return json.dumps(res, indent=4, sort_keys=True, default=str)

    def handler(self, event, context):
        if 'detail-type' in event and event['detail-type'] == 'Scheduled Event':
            res_name = event['resources'][0].split('/')[-1]
            if not res_name.startswith(self.sfn_name):
                raise BadRequestError(f"Biz {self.sfn_name} called with resource {res_name}")

            biz_name = res_name[len(self.sfn_name) + 1:]
            if biz_name not in self.biz:
                raise BadRequestError(f"Unregistered biz : {biz_name}")

            if self.debug:
                print(f"Trigger: {biz_name}")

            return self.biz[biz_name](event, context)

        return super().handler(event, context)


class BizMicroService(TechMicroService):
    """Biz composed microservice activated by a reactor."""

    def __init__(self, biz_factory, data, trigger, **kwargs):
        super().__init__(**kwargs)
        self.biz_factory = biz_factory
        self.data = data
        self.trigger = trigger

    def handler(self, event, context):
        return self.biz_factory.invoke(self.data)


class Blueprint(CoworksMixin, ChaliceBlueprint):
    """Chalice blueprint created directly from class."""

    def __init__(self, **kwargs):
        import_name = kwargs.pop('import_name', self.__class__.__name__)
        super().__init__(import_name)

    @property
    def import_name(self):
        return self._import_name

    @property
    def component_name(self):
        return ''

    @property
    def current_app(self):
        return self._current_app


class Once:
    ...


class At(Cron):
    def __init__(self, minutes=0, hours=1, day_of_month=None, month=None, day_of_week=None, year=None):
        super().__init__(minutes, hours, day_of_month, month, day_of_week, year)

    def to_dict(self):
        return {
            'source': 'at',
            'value': self.to_string(),
        }


class Every(Rate):
    def __init__(self, value, unit):
        super().__init__(value, unit)

    def to_dict(self):
        return {
            'source': 'every',
            'value': self.to_string(),
        }



from dataclasses import dataclass

from chalice import CORSConfig as ChaliceCORSConfig


class CORSConfig(ChaliceCORSConfig):

    def get_access_control_headers(self):
        if not self.allow_origin:
            return {}
        return super().get_access_control_headers()


@dataclass
class Config:
    cors: CORSConfig = CORSConfig(allow_origin='')
    timeout: int = 60



import os
import threading
import time

import requests

from coworks.cws.client import CwsCLIFactory
from coworks.pytest.local_server import ThreadedLocalServer
from .example import TechMS

EXAMPLE_DIR = os.getenv('EXAMPLE_DIR')


def test_simple_example(local_server_factory):
    local_server = local_server_factory(TechMS())
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == "Simple microservice for test.\n"
    response = local_server.make_call(requests.get, '/', params={"usage": "demo"})
    assert response.status_code == 200
    assert response.text == "Simple microservice for demo.\n"
    response = local_server.make_call(requests.get, '/value/1')
    assert response.status_code == 200
    assert response.text == "0\n"
    response = local_server.make_call(requests.put, '/value/1', json={"value": 456})
    assert response.status_code == 200
    assert response.text == "456"
    response = local_server.make_call(requests.get, '/value/1')
    assert response.status_code == 200
    assert response.text == "456\n"


def test_params(local_server_factory):
    local_server = local_server_factory(TechMS())
    response = local_server.make_call(requests.put, '/value/1', json=456)
    assert response.status_code == 200
    assert response.text == "456"
    response = local_server.make_call(requests.get, '/value/1')
    assert response.status_code == 200
    assert response.text == "456\n"


def test_env(local_server_factory):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    local_server = local_server_factory(TechMS(), config_path=dir_path)
    response = local_server.make_call(requests.get, '/env')
    assert response.status_code == 200
    assert response.text == "Simple microservice for test environment variable.\n"


def run_server_example(app, port):
    print(f"Server starting on port {port}")
    app.run(host='localhost', port=port, project_dir=EXAMPLE_DIR)


def test_run_example():
    app = CwsCLIFactory.import_attr('example', 'app', cwd=EXAMPLE_DIR)
    port = ThreadedLocalServer.unused_tcp_port()
    server = threading.Thread(target=run_server_example, args=(app, port), daemon=True)
    server.start()
    counter = 1
    time.sleep(counter)
    while not server.is_alive() and counter < 3:
        time.sleep(counter)
        counter += 1
    response = requests.get(f'http://localhost:{port}/')
    assert response.text == "Simple microservice for test.\n"
    response = requests.get(f'http://localhost:{port}/', params={"usage": "demo"})
    assert response.status_code == 200
    assert response.text == "Simple microservice for demo.\n"
    app.local_server.shutdown()



import os
from collections import defaultdict

from coworks import TechMicroService
from coworks.cws import TerraformWriter


class TechMS(TechMicroService):
    """Technical microservice for the CoWorks tutorial example."""
    version = "1.2"
    values = defaultdict(int)

    def get(self, usage="test"):
        """Entrypoint for testing named parameter."""
        return f"Simple microservice for {usage}.\n"

    def get_value(self, index):
        """Entrypoint for testing positional parameter."""
        return f"{self.values[index]}\n"

    def put_value(self, index, value=0):
        self.values[index] = value
        return value

    def get_env(self):
        return f"Simple microservice for {os.getenv('test')}.\n"


# usefull for test info (don't remove)
tech_app = TechMS()
TerraformWriter(tech_app)

app = TechMS()

if __name__ == '__main__':
    app.run()



import io
import os
import threading
import time

import requests

from coworks.cws.client import CwsCLIFactory, export_to_file
from coworks.pytest.local_server import ThreadedLocalServer

EXAMPLE_DIR = os.getenv('EXAMPLE_DIR')


def run_server_quickstart(app, port):
    print(f"Server starting on port {port}")
    app.run(host='localhost', port=port, project_dir=EXAMPLE_DIR)


def test_run_quickstart():
    app = CwsCLIFactory.import_attr('quickstart2', 'app', cwd=EXAMPLE_DIR)
    port = ThreadedLocalServer.unused_tcp_port()
    server = threading.Thread(target=run_server_quickstart, args=(app, port), daemon=True)
    server.start()
    counter = 1
    time.sleep(counter)
    while not server.is_alive() and counter < 3:
        time.sleep(counter)
        counter += 1
    response = requests.get(f'http://localhost:{port}/', headers={'Authorization': "token"})
    assert response.text == "Simple microservice ready.\n"
    app.local_server.shutdown()


def test_export_quickstart():
    app = CwsCLIFactory.import_attr('quickstart2', 'app', cwd=EXAMPLE_DIR)
    output = io.StringIO()
    export_to_file('quickstart2', 'app', 'terraform', output, project_dir=EXAMPLE_DIR)
    output.seek(0)
    print(output.read())



import os
import pathlib
import re
import sys
from zipfile import ZipFile

from SCons import Environment
from SCons.Builder import Builder


# noinspection PyPep8Naming

def Layer(filename='layer.zip', source_dirs=None, modules=None, exclude_dirs=None, exclude_file_pattern=None):
    env = Environment.Base(BUILDERS={'Layer': Builder(action=generate_zip_file)}, SRC_DIRS=source_dirs,
                           MODULES=modules, EXCLUDE_DIRS=exclude_dirs, EXCLUDE_FILE_PATTERN=exclude_file_pattern)
    env.Layer(filename, [])


def generate_zip_file(target, source, env=None):
    source_dirs = env['SRC_DIRS']
    modules = env['MODULES']
    exclude_dirs = env['EXCLUDE_DIRS']
    exclude_file_pattern = env['EXCLUDE_FILE_PATTERN']

    source_dirs = source_dirs or [path for path in sys.path if os.path.basename(path) == 'site-packages']
    if type(source_dirs) is not list:
        source_dirs = [source_dirs]

    exclude_dirs = exclude_dirs or ['__pycache__', 'pip', 'botocore', 'boto3', 'click', 'docutils', 'tests']
    exclude_file_pattern = exclude_file_pattern or r'.*[.]pyc'
    for source_dir in source_dirs:
        with ZipFile(target[0].name, mode='w') as layer_zip:
            for root, dirs, files in os.walk(source_dir):
                pathes = pathlib.Path(os.path.relpath(root, source_dir))
                if any(part in exclude_dirs for part in pathes.parts):
                    continue

                if modules and root not in modules:
                    continue

                for file in files:
                    if re.match(exclude_file_pattern, file) is not None:
                        continue
                    filename = os.path.join(root, file)
                    if os.path.isfile(filename):  # regular files only
                        arcname = os.path.join('python', os.path.relpath(filename, source_dir))
                        layer_zip.write(filename, arcname)
            for root, dirs, files in os.walk('/home/studiogdo/workspace/coworks/coworks'):
                # add directory (needed for empty dirs)
                # layer_zip.write(root, os.path.relpath(root, relroot))
                for file in files:
                    if re.match(exclude_file_pattern, file) is not None:
                        continue
                    filename = os.path.join(root, file)
                    if os.path.isfile(filename):  # regular files only
                        arcname = os.path.join('python', os.path.relpath(filename, '/home/studiogdo/workspace/coworks'))
                        layer_zip.write(filename, arcname)



import requests
import os

from tests.src.coworks.tech_ms import *

EXAMPLE_DIR = os.getenv('EXAMPLE_DIR')


class WithEnvMS(SimpleMS):

    def get(self):
        """Root access."""
        return os.getenv("test")


def test_default(local_server_factory):
    local_server = local_server_factory(WithEnvMS(), config_path=EXAMPLE_DIR)
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'test environment variable'


def test_dev_stage(local_server_factory):
    local_server = local_server_factory(WithEnvMS(), config_path=EXAMPLE_DIR, stage="dev")
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'test environment variable'


def test_prod_stage(local_server_factory):
    local_server = local_server_factory(WithEnvMS(), config_path=EXAMPLE_DIR, stage="master")
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'prod variable'



from coworks import TechMicroService
from coworks.cws.writer import TerraformWriter


class SimpleMicroService(TechMicroService):

    def get(self):
        return f"Simple microservice ready.\n"

    def auth(self, auth_request):
        return auth_request.token == "token"


app = SimpleMicroService(app_name='test')
TerraformWriter(app)



import json
import os

import pytest
from chalice.config import Config
from chalice.cli import DEFAULT_STAGE_NAME

from .local_server import ThreadedLocalServer


@pytest.fixture()
def local_server_factory():
    threaded_server = ThreadedLocalServer()

    def create_server(app, **kwargs):

        # if config_path defined, use it to update environment from conf json file
        if 'config_path' in kwargs:
            config_path = kwargs.pop('config_path')
            config_file = os.path.join(config_path, '.chalice', 'config.json')
            with open(config_file) as f:
                config_from_disk = json.loads(f.read())
            chalice_stage = kwargs.pop('stage') if 'stage' in kwargs else DEFAULT_STAGE_NAME
            config = Config(config_from_disk=config_from_disk, chalice_stage=chalice_stage)
            os.environ.update(config.environment_variables)
            kwargs['config'] = config

        threaded_server.configure(app, **kwargs)
        threaded_server.start()
        return threaded_server

    try:
        yield create_server
    finally:
        threaded_server.shutdown()



import json
import os
import traceback

import boto3
import requests
from chalice import Chalice, BadRequestError

app = Chalice(app_name="app")


@app.lambda_function(name="call")
def call_url(event, context):
    from aws_xray_sdk.core.lambda_launcher import LAMBDA_TRACE_HEADER_KEY
    from aws_xray_sdk.core.models.trace_header import TraceHeader
    from aws_xray_sdk.core.context import Context
    from aws_xray_sdk.core import xray_recorder, patch

    libraries = (['botocore', 'requests'])
    patch(libraries)

    # Set X-Ray trace data
    x_ray = event.get('XRay')
    ms_name = event.get('MSName', 'no name')
    if x_ray:
        root_id = x_ray.get('Root')
        parent_id = x_ray.get('Parent')
        trace_header = TraceHeader(root=root_id, parent=parent_id)
    else:
        header_str = os.getenv(LAMBDA_TRACE_HEADER_KEY)
        trace_header = TraceHeader.from_header_str(header_str)
        root_id = trace_header.root
        parent_id = trace_header.parent

    # XRay segment
    xray_recorder.configure(service='coworks-step-function', context=Context())
    segment = xray_recorder.begin_segment(f"{ms_name} Step Function call", traceid=root_id, parent_id=parent_id)
    try:
        # Get invoke parameters
        environment = event.get('environment')
        invoke_params = event.get('invoke_params')
        query_params = invoke_params.get('query_params', {})
        body_params = invoke_params.get('body_params', {})
        headers = invoke_params.get('headers', {})
        inputs = event.get('inputs')
        output = event.get('output')

        # Get S3 streamed parameters
        kwargs = {}
        if inputs:
            # subsegment = xray_recorder.begin_segment('Get streamed parameters')
            for input in inputs:
                aws_access_key_id = environment.get('aws_access_key_id', {})
                aws_secret_access_key = environment.get('aws_secret_access_key', {})
                region_name = environment.get('region_name', {})
                bucket = input.get('bucket')
                key = input.get('key')
                boto_session = boto3.Session(aws_access_key_id, aws_secret_access_key,
                                             region_name=region_name)
                s3_client = boto_session.client('s3')
                uploaded_object = s3_client.get_object(Bucket=bucket, Key=key)
                if not uploaded_object:
                    return BadRequestError(f"No key '{key}' in bucket '{bucket}'")
                value = uploaded_object['Body'].read().decode('utf-8')
                print(value)
                print(type(value))
                kwargs[input['name']] = value
            # subsegment.put_metadata('streams', kwargs.keys(), "my namespace")
            # xray_recorder.end_segment()

        # Calls APIGateway url
        http_method = invoke_params.get('http_method')
        if not http_method:
            raise BadRequestError("http method parameter not defined")
        method = getattr(requests, http_method.lower())
        if not method:
            raise BadRequestError(f"Undefined http method {http_method}")
        url = invoke_params.get('url')
        if not url:
            raise BadRequestError("url method parameter not defined")

        if query_params:
            for name, content in kwargs.items():
                query_params[name] = content
        if body_params:
            for name, content in kwargs.items():
                body_params[name] = content
            # use JSONPath to change value

        # Uses same trace id of the first call, set metadata and invko microservice
        headers['X-Amzn-Trace-Id'] = trace_header.to_header_str()
        for key, value in event.items():
            segment.put_metadata(key, value, ms_name)
        resp = _call_url(method, url, params=query_params, json=body_params, headers=headers)
        segment.put_metadata('call_result', resp.text, ms_name)
        xray = {'root': root_id, 'parent': segment.id}
        resp.raise_for_status()

        # Stores S3 streamed parameters
        if output:
            # subsegment = xray_recorder.begin_subsegment('Put streamed output')
            aws_access_key_id = environment.get('aws_access_key_id', {})
            aws_secret_access_key = environment.get('aws_secret_access_key', {})
            region_name = environment.get('region_name', {})
            bucket = output.get('bucket', "morassuti-sfn")
            key = output.get('key', "stock_armony")
            boto_session = boto3.Session(aws_access_key_id, aws_secret_access_key,
                                         region_name=region_name)
            s3_client = boto_session.client('s3')
            s3_client.put_object(Bucket=bucket, Key=key, Body=resp.text)
            # subsegment.put_metadata("Output stream", {'Bucket': bucket, 'Key': key}, "my namespace")
            # xray_recorder.end_subsegment()

            result = {'bucket': bucket, 'key': key}
            return {'result': result, 'xray': xray}

        try:
            return {'result': resp.json(), 'xray': xray}
        except json.JSONDecodeError:
            return resp.content
    except Exception as e:
        print(f"Exception : {str(e)}")
        traceback.print_exc()
        segment.add_exception(e, traceback.extract_stack())
        raise
    finally:
        xray_recorder.end_segment()


def _call_url(method, url, params, json, headers):
    if not url:
        raise BadRequestError(f"Undefined url {url}.")

    if params:
        return method(url, params=params, headers=headers)
    if json:
        return method(url, json=json, headers=headers)
    return method(url, headers=headers)



from dataclasses import dataclass
from typing import List

from coworks import TechMicroService
from coworks.config import CORSConfig
from .writer import TemplateWriter

UID_SEP = '_'


@dataclass
class Entry:
    app: TechMicroService
    parent_uid: str
    path: str
    methods: List[str]
    cors: CORSConfig

    @property
    def uid(self):
        def remove_brackets(path):
            return f"{path.replace('{', '').replace('}', '')}"

        if self.path is None:
            return UID_SEP

        last = remove_brackets(self.path)
        return f"{self.parent_uid}{UID_SEP}{last}" if self.parent_uid else last

    @property
    def is_root(self):
        return self.path is None

    @property
    def parent_is_root(self):
        return self.parent_uid == UID_SEP

    def __repr__(self):
        return f"{self.uid}:{self.methods}"


class TerraformWriter(TemplateWriter):

    def __init__(self, app=None, name='terraform', data=None, **kwargs):
        data = data or {
            'layer_zip_file': 'layer.zip',
        }
        super().__init__(app=app, name=name, data=data, **kwargs)

    @property
    def default_template_filenames(self):
        return ['terraform.j2']

    def _export_header(self, **kwargs):
        print("// Do NOT edit this file as it is auto-generated by cws\n", file=self.output)

    @property
    def entries(self):
        """Returns the list of flatten path (prev, last, keys)."""
        all_pathes_id = {}

        def add_entry(previous, last, meth):
            entry = Entry(self.app, previous, last, meth, self.app.config.cors)
            uid = entry.uid
            if uid not in all_pathes_id:
                all_pathes_id[uid] = entry
            if all_pathes_id[uid].methods is None:
                all_pathes_id[uid].methods = meth
            return uid

        for route, methods in self.app.routes.items():
            previous_uid = UID_SEP
            splited_route = route[1:].split('/')

            # special root case
            if splited_route == ['']:
                add_entry(None, None, methods.keys())
                continue

            # creates intermediate resources
            last_path = splited_route[-1:][0]
            for prev in splited_route[:-1]:
                previous_uid = add_entry(previous_uid, prev, None)

            # set entryes keys for last entry
            add_entry(previous_uid, last_path, methods.keys())

        return all_pathes_id


class TerraformPlanWriter(TerraformWriter):

    def __init__(self, app=None, name='plan', data=None, template_filenames=None, **kwargs):
        template_filenames = template_filenames or ['plan.j2']
        data = data or {'cmd': 'plan'}
        super().__init__(app, name=name, template_filenames=template_filenames, data=data, **kwargs)

    def _export_header(self, **kwargs):
        pass

    def _format(self, content):
        return content.translate(str.maketrans('\n', ' ', '\t\r'))



import io
import json
from unittest.mock import Mock, MagicMock

import pytest
import yaml

from coworks import BizFactory
from coworks.cws import WriterError
from coworks.cws.writer.sfn import StepFunctionWriter, StepFunction, TechState
from tests.src.coworks.tech_ms import S3MockTechMS


class TestStepFunction(StepFunction):

    def __init__(self, data):
        filepath = Mock()
        file = MagicMock()
        file.__enter__ = Mock(return_value=data)
        filepath.open = Mock(return_value=file)
        with filepath.open() as file:
            self.data = yaml.load(file, Loader=yaml.SafeLoader)

        super().__init__("test", filepath)


class TechMS(S3MockTechMS):

    def get_test(self):
        return "get"

    def get_params(self, value, other):
        return f"get {value} and {other}"

    def get_params_(self, value=1, other=2):
        return f"get {value} and {other}"

    def post_params(self, value=1, other=2):
        return f"get {value} and {other}"


def test_no_params():
    tech = TechMS()

    data = {'get': '/test'}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200


def test_arg_params():
    tech = TechMS()

    uri_params = {'_0': 1, '_1': 2}
    data = {'get': '/params/{_0}/{_1}', 'uri_params': uri_params}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200
    assert res['body'] == "get 1 and 2"


def test_kwargs_params():
    tech = TechMS()

    data = {'get': '/params'}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200
    assert res['body'] == "get 1 and 2"

    query_params = {'value': [3], 'other': [4]}
    data = {'get': '/params', 'query_params': query_params}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200
    assert res['body'] == "get 3 and 4"

    query_params = {'value': [5], 'other': [6]}
    data = {'post': '/params', 'query_params': query_params}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200
    assert res['body'] == "get 5 and 6"

    body = {'value': 7, 'other': 8}
    data = {'post': '/params', 'body': body}
    call = TechState.get_call_data(None, data)
    res = tech(call, {})
    assert res['statusCode'] == 200
    assert res['body'] == "get 7 and 8"


def test_biz_empty():
    biz = BizFactory('tests/src/coworks/biz/empty')
    biz.create('test')
    writer = StepFunctionWriter(biz)
    output = io.StringIO()
    with pytest.raises(WriterError):
        writer.export(output=output, error=output)
    output.seek(0)
    res = output.read()
    assert res == "Error in tests/src/coworks/biz/empty: The content of the tests/src/coworks/biz/empty microservice " \
                  "seems to be empty.\n"


def test_biz_complete():
    """Tests the doc example."""
    fact = BizFactory('tests/src/coworks/biz/complete')
    fact.create('test')
    writer = StepFunctionWriter(fact)
    output = io.StringIO()
    writer.export(output=output, error=output)
    output.seek(0)
    source = json.loads(output.read())
    assert source['Version'] == "1.0"
    assert 'Comment' in source
    assert len(source['States']) == 4

    states = source['States']
    state = states['Check server']
    assert state is not None
    assert state['Type'] == 'Task'

    state = states['Send mail']
    assert state is not None
    assert state['Type'] == 'Task'
    assert state['End'] is True


def test_fail():
    data = {'states': [{
        'name': "fail",
        'fail': None
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert len(sfn.all_states) == 2
    assert 'End' not in sfn.all_states[1].state
    assert 'Cause' in sfn.all_states[1].state
    assert 'Error' in sfn.all_states[1].state


def test_pass():
    # missing key cases
    data = {'states': [{
        'name': "action",
        'pass': None
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert len(sfn.all_states) == 2
    assert 'End' in sfn.all_states[1].state


def test_tech():
    # missing key cases
    data = {'states': [{
        'name': "action",
        'tech': {
            'service': "tech",
        }
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    with pytest.raises(WriterError) as pytest_wrapped_e:
        sfn.generate()
    assert pytest_wrapped_e.type == WriterError
    assert pytest_wrapped_e.value.args[0] == "No route defined for {'name': 'action', 'tech': {'service': 'tech'}}"

    data = {'states': [{
        'name': "action",
        'tech': {
            'get': "/",
        }
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    with pytest.raises(WriterError) as pytest_wrapped_e:
        sfn.generate()
    assert pytest_wrapped_e.type == WriterError
    assert pytest_wrapped_e.value.args[0] == "The key service is missing for {'get': '/'}"

    data = {'states': [{
        'name': "action",
        'tech': {
            'service': "tech1",
            'get': "/",
            'post': "/",
        }
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    with pytest.raises(WriterError) as pytest_wrapped_e:
        sfn.generate()
    assert pytest_wrapped_e.type == WriterError
    assert pytest_wrapped_e.value.args[0].startswith("Too many methods defined for ")


def test_catch_all():
    data = {'states': [{
        'name': "action",
        'tech': {
            'service': "tech1",
            'get': "/",
        }
    }], 'catch': [{'fail': None}]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert len(sfn.all_states) == 4
    assert len(sfn.all_states[1].state['Catch']) == 1

    data = {'states': [{
        'name': "action",
        'tech': {
            'service': "tech1",
            'get': "/",
        },
        'catch': [{'fail': None}]
    }], 'catch': [{'fail': None}]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert len(sfn.all_states) == 5
    assert len(sfn.all_states[1].state['Catch']) == 2


def test_list():
    data = {'states': [{
        'name': "action 1",
        'tech': {
            'service': "tech",
            'get': "/",
        }
    }, {
        'name': "action 2",
        'tech': {
            'service': "tech",
            'get': "/",
        }
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert 'Next' in sfn.all_states[1].state
    assert sfn.all_states[1].state['Next'] == sfn.all_states[2].name
    assert 'End' in sfn.all_states[2].state


def test_choice():
    # missing key cases
    data = {'states': [{
        'name': "action",
        'choices': None
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    with pytest.raises(WriterError):
        sfn.generate()

    data = {'states': [{
        'name': "action",
        'choices': [{
            'var': '', 'oper': '', 'value': '', 'goto': 'Step'
        }]
    }, {
        'name': "Step",
        'pass': None
    }]}
    sfn = TestStepFunction(yaml.dump(data))
    sfn.generate()
    assert len(sfn.all_states) == 3
    assert 'Default' in sfn.all_states[1].state
    assert 'End' in sfn.all_states[2].state



import io
import json
import os
import shutil

import pytest

from coworks.cws.client import client

EXAMPLE_DIR = os.getenv('EXAMPLE_DIR')


def test_init():
    chalice_dir = os.path.join('test', '.chalice')
    if os.path.exists(chalice_dir):
        shutil.rmtree(chalice_dir)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', 'test', 'init'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    assert os.path.exists(chalice_dir)
    assert os.path.exists(f"{chalice_dir}/config.json")

    shutil.rmtree(chalice_dir)


def test_info():
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'info'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'info', '-m', 'quickstart2'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        output = io.StringIO()
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'info', '-m', 'quickstart2', '-o', output], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0
    output.seek(0)
    out = json.loads(output.read())
    assert 'name' in out
    assert out['name'] == "test"
    assert 'type' in out
    assert out['type'] == "tech"

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'info', '-m', 'example', '-a', 'test'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'info', '-m', 'example', '-a', 'tech_app'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_export():
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        client(prog_name='cws', args=['-p', EXAMPLE_DIR, 'export', '-m', 'example', '-a', 'tech_app'], obj={})
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0



import json
import os
import shutil
import sys
import traceback
from tempfile import SpooledTemporaryFile

import click
from chalice import BadRequestError
from chalice.cli import CONFIG_VERSION, DEFAULT_STAGE_NAME, DEFAULT_APIGATEWAY_STAGE_NAME
from chalice.cli import chalice_version, get_system_info
from chalice.local import LocalChalice
from chalice.utils import serialize_to_json
from coworks import TechMicroService, BizMicroService, BizFactory
from coworks.cws.writer import Writer
from coworks.version import __version__

from .factory import CwsCLIFactory


class CLIError(Exception):
    ...


@click.group()
@click.version_option(version=__version__,
                      message=f'%(prog)s %(version)s, chalice {chalice_version}, {get_system_info()}')
@click.option('-p', '--project-dir',
              help='The project directory path (absolute or relative). Defaults to CWD')
@click.pass_context
def client(ctx, project_dir=None):
    if project_dir is None:
        project_dir = os.getcwd()
    elif not os.path.isabs(project_dir):
        project_dir = os.path.abspath(project_dir)
    ctx.obj['project_dir'] = project_dir


@client.command('init')
@click.option('--force/--no-force', default=False,
              help='Forces project reinitialization.')
@click.pass_context
def init(ctx, force):
    """Init chalice configuration file."""
    project_name = os.path.basename(os.path.normpath(ctx.obj['project_dir']))

    chalice_dir = os.path.join(ctx.obj['project_dir'], '.chalice')
    if os.path.exists(chalice_dir):
        if force:
            shutil.rmtree(chalice_dir)
            created = False
        else:
            sys.stderr.write(f"Project {project_name} already initialized\n")
            return
    else:
        created = True

    os.makedirs(chalice_dir)
    config = os.path.join(chalice_dir, 'config.json')
    cfg = {
        'version': CONFIG_VERSION,
        'app_name': project_name,
        'stages': {
            DEFAULT_STAGE_NAME: {
                'api_gateway_stage': DEFAULT_APIGATEWAY_STAGE_NAME,
            }
        }
    }
    with open(config, 'w') as f:
        f.write(serialize_to_json(cfg))

    if created:
        sys.stdout.write(f"Project {project_name} initialized\n")
    else:
        sys.stdout.write(f"Project {project_name} reinitialized\n")


@client.command('info')
@click.option('-m', '--module', default='app',
              help="Filename of your microservice python source file.")
@click.option('-a', '--app', default='app',
              help="Coworks application in the source file.")
@click.option('-o', '--out')
@click.pass_context
def info(ctx, module, app, out):
    """Information on a microservice."""
    try:
        handler = import_attr(module, app, cwd=ctx.obj['project_dir'])
        if out is not None:
            soutput = open(out, 'w+') if type(out) is str else out
        else:
            out = sys.stdout
        print(json.dumps({
            'name': handler.app_name,
            'type': handler.ms_type
        }), file=out)
    except CLIError:
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(1)


@client.command('run')
@click.option('-m', '--module', default='app',
              help="Filename of your microservice python source file.")
@click.option('-a', '--app', default='app',
              help="Coworks application in the source file.")
@click.option('-h', '--host', default='127.0.0.1')
@click.option('-p', '--port', default=8000, type=click.INT)
@click.option('-s', '--stage', default='dev')
@click.option('--debug/--no-debug', default=False,
              help='Print debug logs to stderr.')
@click.pass_context
def run(ctx, module, app, host, port, stage, debug):
    """Run local server."""
    try:
        handler = import_attr(module, app, cwd=ctx.obj['project_dir'])
        handler.run(host=host, port=port, stage=stage, debug=debug, project_dir=ctx.obj['project_dir'])
    except CLIError:
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(1)


@client.command('export')
@click.option('-m', '--module', default='app',
              help="Filename of your microservice python source file.")
@click.option('-a', '--app', default='app',
              help="Coworks application in the source file.")
@click.option('-b', '--biz', default=None,
              help="BizMicroservice name.")
@click.option('-f', '--format', default='terraform')
@click.option('-o', '--out')
@click.option('--debug/--no-debug', default=False,
              help='Print debug logs to stderr.')
@click.pass_context
def export(ctx, module, app, biz, format, out, debug):
    """Export microservice in other description languages."""
    try:
        export_to_file(module, app, format, out, project_dir=ctx.obj['project_dir'], biz=biz)
    except CLIError:
        sys.exit(1)
    except Exception as e:
        if debug:
            traceback.print_exc()
        sys.stderr.write(str(e))
        sys.exit(1)


@client.command('update')
@click.option('-m', '--module', default='app',
              help="Filename of your microservice python source file.")
@click.option('-a', '--app', default='app',
              help="Coworks application in the source file.")
@click.option('-p', '--profile', default=None)
@click.pass_context
def update(ctx, module, app, profile):
    """Update biz microservice."""
    out = SpooledTemporaryFile(mode='w+')
    handler = export_to_file(module, app, 'sfn', out, project_dir=ctx.obj['project_dir'])
    if handler is None:
        sys.exit(1)
    handler.aws_profile = profile

    out.seek(0)
    if isinstance(handler, BizFactory):
        try:
            update_sfn(handler, out.read())
        except BadRequestError:
            sys.stderr.write(f"Cannot update undefined step function '{handler.sfn_name}'.\n")
            sys.exit(1)

    else:
        sys.stderr.write(f"Update not defined on {type(handler)}.\n")
        sys.exit(1)


def import_attr(module, app, cwd):
    try:
        return CwsCLIFactory.import_attr(module, app, cwd=cwd)
    except AttributeError:
        sys.stderr.write(f"Module '{module}' has no microservice {app} : {str(e)}\n")
        raise CLIError()
    except ModuleNotFoundError as e:
        sys.stderr.write(f"They module '{module}' is not defined in {cwd} : {str(e)}\n")
        raise CLIError()
    except Exception as e:
        sys.stderr.write(f"Error {e} when loading module '{module}'\n")
        raise CLIError()


def export_to_file(module, app, _format, out, **kwargs):
    handler = import_attr(module, app, cwd=kwargs['project_dir'])

    try:
        _writer: Writer = handler.extensions['writers'][_format]
    except KeyError as e:
        sys.stderr.write(f"Format '{_format}' undefined (you haven't add a {_format} writer to {app} )\n")
        raise CLIError()

    _writer.export(output=out, module_name=module, handler_name=app, **kwargs)


def update_sfn(handler, src):
    sfn_client = handler.sfn_client
    response = sfn_client.update_state_machine(stateMachineArn=handler.sfn_arn, definition=src)
    print(response)


def main():
    return client(obj={})



import io
from unittest.mock import MagicMock

from chalice import Response

from coworks import TechMicroService


class TechMS(TechMicroService):
    def __init__(self):
        super().__init__(app_name='test')


class S3MockTechMS(TechMicroService):
    def __init__(self):
        super().__init__(app_name='test')
        session = MagicMock()
        session.client = MagicMock()
        s3_object = {'Body': io.BytesIO(b'test'), 'ContentType': 'text/plain'}
        session.client.get_object = MagicMock(return_value=s3_object)
        self.aws_s3_run_session = session


class SimpleMS(TechMS):

    def get(self):
        """Root access."""
        return "get"

    def get1(self):
        """Not recognized."""
        return "get1"

    def get_content(self):
        return "get content"

    def get__content(self, value):
        return f"get content with {value}"

    def get_content__(self, value, other):
        return f"get content with {value} and {other}"

    def post_content(self, other="none"):
        return f"post content without value but {other}"

    def post_content_(self, value, other="none"):
        return f"post content with {value} and {other}"

    # **param
    def get_kwparam1(self, value=0):
        return f"get **param with only {value}"

    def get_kwparam2(self, value=0, **kwargs):
        return f"get **param with {value} and {list(kwargs.keys())}"

    # composed path
    def get_extended_content(self):
        return "hello world"


class PrefixedMS(TechMS):
    url_prefix = 'prefix'

    def get(self):
        return "hello world"

    def get_content(self):
        return "hello world"

    def get_extended_content(self):
        return "hello world"


class ParamMS(TechMS):
    value = "123"

    def get(self, str):
        return str

    def get_concat(self, str1, str2):
        return str1 + str2

    def get_value(self):
        return self.value

    def put_value(self):
        request = self.current_request
        self.value = request.json_body['value']
        return self.value

    def get_param(self, str1, param1='default1', param2='default2'):
        return str1 + str(param1) + param2


class PrefixedParamMS(TechMS):
    url_prefix = 'prefix'

    def get(self, str):
        return str

    def get_concat(self, str1, str2):
        return str1 + str2


class TupleReturnedMS(TechMS):
    def get(self):
        return 'ok', 200

    def get_resp(self, str):
        return Response(body=str, status_code=200)

    def get_error(self, str):
        return str, 300



import requests
import json
import os
import urllib.parse
from coworks.tech import JinjaRenderMicroService


def test_render_template_in_url(local_server_factory):
    local_server = local_server_factory(JinjaRenderMicroService())
    template = urllib.parse.quote_plus("hello {{ world_name }}")
    response = local_server.make_call(requests.get, f"/render/{template}", params={'world_name': 'world'})
    assert response.status_code == 200
    assert response.json() == {"render": "hello [\'world\']"}


def test_render_template_multipart_form(local_server_factory):
    local_server = local_server_factory(JinjaRenderMicroService())

    template = open("template.jinja", "w+")
    template.write("hello {{ world_name }}")
    template.close()

    context = open("context.json", "w+")
    context.write(json.dumps({"world_name": "the world"}))
    context.close()

    template = open("template.jinja", "r")
    context = open("context.json", "r")

    response = local_server.make_call(requests.post, f"/render/template.jinja",
                                      files={'templates': ('template.jinja', template, 'text/plain'),
                                             'context': (None, context, 'application/json')})
    assert response.status_code == 200
    assert response.json() == {"render": "hello the world"}
    os.remove("template.jinja")



import io
import json
from unittest.mock import MagicMock

import requests

from coworks import TechMicroService
from coworks.cws.writer.sfn import TechState


class TechMS(TechMicroService):
    def __init__(self):
        super().__init__(app_name='test')

    def post_params(self, text=None, context=None, files=None):
        if files:
            if type(files) is not list:
                files = [files]
            return f"post {text}, {context} and {[f.file.name for f in files]}"
        return f"post {text}, {context}"


session = MagicMock()
session.client = MagicMock()
s3_object = {'Body': io.BytesIO(b'test'), 'ContentType': 'text/plain'}
session.client.get_object = MagicMock(return_value=s3_object)


def test_text_plain(local_server_factory):
    """normal API call."""
    tech = TechMS()
    local_server = local_server_factory(tech)
    data = json.dumps({'text': 'value'})
    headers = {'Content-type': 'text/plain'}
    response = local_server.make_call(requests.post, '/params', data=data, timeout=500, headers=headers)
    assert response.status_code == 415


def test_form_data(local_server_factory):
    """normal API call."""
    tech = TechMS()
    tech.aws_s3_form_data_session = session
    local_server = local_server_factory(tech)
    data = {'key': 'value'}
    multiple_files = [
        ('text', (None, "hello world")),
        ('context', (None, json.dumps(data), 'application/json')),
        ('files', ('f1.csv', 'some,data,to,send\nanother,row,to,send\n')),
        ('files', ('f2.txt', 'some,data,to,send\nanother,row,to,send\n', 'text/plain')),
        ('files', ('f3.j2', 'bucket/key', 'text/s3')),
    ]
    response = local_server.make_call(requests.post, '/params', files=multiple_files, json=data)
    assert response.status_code == 200
    assert response.text == "post hello world, {'key': 'value'} and ['f1.csv', 'f2.txt', 'f3.j2']"


def test_sfn_arg_params():
    """step function call."""
    tech = TechMS()
    tech.aws_s3_form_data_session = session
    form_data = {
        'text': {'content': "hello world"},
        'context': {'content': {'key': 'value'}, 'mime_type': 'application/json'},
        'files': [
            {'filename': 'f1.csv', 'content': 'some,data', 'mime_type': 'text/plain'},
            {'filename': 'f2.txt', 'content': 'content2', 'mime_type': 'text/plain'},
            {'filename': 'f3.j2', 'path': 'bucket/key', 'mime_type': 'text/s3'},
        ],
    }
    data = {'post': '/params', 'form-data': form_data}
    call = TechState.get_call_data(None, data)
    response = tech(call, {})
    assert response['statusCode'] == 200
    assert response['body'] == "post hello world, {'key': 'value'} and ['f1.csv', 'f2.txt', 'f3.j2']"


def test_sfn_simple_arg_params():
    """simplified step function call."""
    tech = TechMS()
    tech.aws_s3_form_data_session = session
    form_data = {
        'text': "hello world",
        'context': {'json': True},
        'files': {'s3': 'bucket/key'},
    }
    data = {'post': '/params', 'form-data': form_data}
    call = TechState.get_call_data(None, data)
    response = tech(call, {})
    assert response['statusCode'] == 200
    assert response['body'] == "post hello world, True and ['key']"



import os
import smtplib
from email.message import EmailMessage

from aws_xray_sdk.core import xray_recorder
from chalice import ChaliceViewError, BadRequestError

from coworks import TechMicroService, FileParam


class MailMicroService(TechMicroService):
    """Mail microservice"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.smtp_server = self.smtp_login = self.smtp_passwd = None

        @self.before_first_request
        def check_env_vars():
            self.smtp_server = os.getenv('SMTP_SERVER')
            if not self.smtp_server:
                raise EnvironmentError('SMTP_SERVER not defined in environment')
            self.smtp_login = os.getenv('SMTP_LOGIN')
            if not self.smtp_login:
                raise EnvironmentError('SMTP_LOGIN not defined in environment')
            self.smtp_passwd = os.getenv('SMTP_PASSWD')
            if not self.smtp_passwd:
                raise EnvironmentError('SMTP_PASSWD not defined in environment')

    def post_send(self, subject="", from_addr: str = None, to_addrs: [str] = None, body="",
                  attachments: [FileParam] = None, subtype="plain", starttls=True):
        """ Send mail.
        To send attachments, add files in the body of the request as multipart/form-data. """

        from_addr = from_addr or os.getenv('from_addr')
        if not from_addr:
            raise BadRequestError("From address not defined (from_addr:str)")
        to_addrs = to_addrs or os.getenv('to_addrs')
        if not to_addrs:
            raise BadRequestError("To addresses not defined (to_addrs:[str])")

        # Creates email
        try:
            msg = EmailMessage()
            msg['Subject'] = subject
            msg['From'] = from_addr
            msg['To'] = to_addrs if isinstance(to_addrs, str) else ', '.join(to_addrs)
            msg.set_content(body, subtype=subtype)

            if attachments:
                if not isinstance(attachments, list):
                    attachments = [attachments]
                for attachment in attachments:
                    if not attachment.mime_type:
                        raise BadRequestError(f"Mime type of the attachment {attachment.file.name} is not defined")
                    maintype, subtype = attachment.mime_type.split("/")
                    msg.add_attachment(attachment.file.read(), maintype=maintype, subtype=subtype,
                                       filename=attachment.file.name)

        except Exception as e:
            raise ChaliceViewError(f"Cannot create email message (Error: {str(e)}).")

        # Send emails
        try:
            with smtplib.SMTP(self.smtp_server) as server:
                if starttls:
                    server.starttls()
                server.login(self.smtp_login, self.smtp_passwd)
                subsegment = xray_recorder.begin_subsegment(f"SMTP sending")
                try:
                    if subsegment:
                        subsegment.put_metadata('message', msg.as_string())
                    server.send_message(msg)
                finally:
                    xray_recorder.end_subsegment()

            return f"Mail sent to {msg['To']}"
        except smtplib.SMTPAuthenticationError:
            raise BadRequestError("Wrong username/password.")
        except Exception as e:
            raise ChaliceViewError(f"Cannot send email message (Error: {str(e)}).")



import requests

from tests.src.coworks.tech_ms import SimpleMS
from tests.src.coworks.blueprint.blueprint import BP


def test_request(local_server_factory):
    ms = SimpleMS()
    ms.register_blueprint(BP(import_name="blueprint"))
    local_server = local_server_factory(ms)
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'get'
    response = local_server.make_call(requests.get, '/test/3')
    assert response.status_code == 200
    assert response.text == 'blueprint test 3'
    response = local_server.make_call(requests.get, '/extended/test/3')
    assert response.status_code == 200
    assert response.text == 'blueprint extended test 3'


def test_prefix(local_server_factory):
    ms = SimpleMS()
    ms.register_blueprint(BP(), url_prefix="/prefix")
    local_server = local_server_factory(ms)
    response = local_server.make_call(requests.get, '/prefix/test/3')
    assert response.status_code == 200
    assert response.text == 'blueprint test 3'
    response = local_server.make_call(requests.get, '/prefix/extended/test/3')
    assert response.status_code == 200
    assert response.text == 'blueprint extended test 3'



import requests

from tests.src.coworks.tech_ms import *


def test_request_arg(local_server_factory):
    local_server = local_server_factory(SimpleMS())
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'get'
    response = local_server.make_call(requests.post, '/')
    assert response.status_code == 405
    response = local_server.make_call(requests.get, '/get1')
    assert response.status_code == 403
    response = local_server.make_call(requests.get, '/content')
    assert response.status_code == 200
    assert response.text == "get content"
    response = local_server.make_call(requests.get, '/content/3')
    assert response.status_code == 200
    assert response.text == "get content with 3"
    response = local_server.make_call(requests.get, '/content/3/other')
    assert response.status_code == 200
    assert response.text == "get content with 3 and other"
    response = local_server.make_call(requests.post, '/content', json={"other": 'other'})
    assert response.status_code == 200
    assert response.text == "post content without value but other"
    response = local_server.make_call(requests.post, '/content/3', json={"other": 'other'})
    assert response.status_code == 200
    assert response.text == "post content with 3 and other"
    response = local_server.make_call(requests.post, '/content/3', json='other')
    assert response.status_code == 200
    assert response.text == "post content with 3 and other"
    response = local_server.make_call(requests.post, '/content/3', json={"other": 'other', "value": 5})
    assert response.status_code == 400


def test_request_kwargs(local_server_factory):
    local_server = local_server_factory(SimpleMS())
    response = local_server.make_call(requests.get, '/kwparam1', params={"value": 5})
    assert response.status_code == 200
    assert response.text == "get **param with only 5"
    response = local_server.make_call(requests.get, '/kwparam1', params={"other": 'other', "value": 5})
    assert response.status_code == 400
    response = local_server.make_call(requests.get, '/kwparam1', json={"value": 5})
    assert response.status_code == 200
    assert response.text == "get **param with only 5"
    response = local_server.make_call(requests.get, '/kwparam1', json={"other": 'other', "value": 5})
    assert response.status_code == 400
    response = local_server.make_call(requests.get, '/kwparam2', params={"other": 'other', "value": 5})
    assert response.status_code == 200
    assert response.text == "get **param with 5 and ['other']"
    response = local_server.make_call(requests.get, '/kwparam2', json={"other": 'other', "value": 5})
    assert response.status_code == 200
    assert response.text == "get **param with 5 and ['other']"

    response = local_server.make_call(requests.get, '/extended/content')
    assert response.status_code == 200
    assert response.text == "hello world"


def test_prefixed(local_server_factory):
    local_server = local_server_factory(PrefixedMS())
    response = local_server.make_call(requests.get, '/prefix')
    assert response.status_code == 200
    assert response.text == "hello world"
    response = local_server.make_call(requests.get, '/prefix/content')
    assert response.status_code == 200
    assert response.text == "hello world"
    response = local_server.make_call(requests.get, '/prefix/extended/content')
    assert response.status_code == 200
    assert response.text == "hello world"


def test_parameterized(local_server_factory):
    local_server = local_server_factory(ParamMS())
    response = local_server.make_call(requests.get, '/123')
    assert response.status_code == 200
    assert response.text == '123'
    response = local_server.make_call(requests.get, '/concat/123/456')
    assert response.status_code == 200
    assert response.text == '123456'
    response = local_server.make_call(requests.get, '/value')
    assert response.status_code == 200
    assert response.text == '123'
    response = local_server.make_call(requests.put, '/value', json={'value': "456"})
    assert response.status_code == 200
    assert response.text == '456'
    response = local_server.make_call(requests.get, '/value')
    assert response.status_code == 200
    assert response.text == '456'
    response = local_server.make_call(requests.get, '/param/test1')
    assert response.status_code == 200
    assert response.text == 'test1default1default2'
    response = local_server.make_call(requests.get, '/param/test1', params={"param1": "param1"})
    assert response.status_code == 200
    assert response.text == 'test1param1default2'
    response = local_server.make_call(requests.get, '/param/test1', params={"param2": "param2"})
    assert response.status_code == 200
    assert response.text == 'test1default1param2'
    response = local_server.make_call(requests.get, '/param/test1', params={"param1": "param1", "param2": "param2"})
    assert response.status_code == 200
    assert response.text == 'test1param1param2'
    response = local_server.make_call(requests.get, '/param/test1', params=[('param1', 'param1'), ('param1', 'param2')])
    assert response.status_code == 200
    assert response.text == "test1['param1', 'param2']default2"


def test_slug_parameterized(local_server_factory):
    local_server = local_server_factory(PrefixedParamMS())
    response = local_server.make_call(requests.get, '/prefix/123')
    assert response.status_code == 200
    assert response.text == '123'
    response = local_server.make_call(requests.get, '/prefix/concat/123/456')
    assert response.status_code == 200
    assert response.text == '123456'


def test_tuple_returned(local_server_factory):
    local_server = local_server_factory(TupleReturnedMS())
    response = local_server.make_call(requests.get, '/')
    assert response.status_code == 200
    assert response.text == 'ok'
    response = local_server.make_call(requests.get, '/resp/ok')
    assert response.status_code == 200
    assert response.text == 'ok'
    response = local_server.make_call(requests.get, '/error/test')
    assert response.status_code == 300
    assert response.text == 'test'


def test_after_before_request():
    simple = SimpleMS()
    evts = []

    @simple.before_first_activation
    def always_first():
        evts.append("first")

    @simple.before_activation
    def before():
        evts.append("before")

    @simple.after_activation
    def always_first():
        evts.append("after")

    simple({}, {})
    simple({}, {})

    assert evts == ["first", "before", "after", "before", "after"]



from unittest.mock import MagicMock
import pytest
import requests
import os

from coworks.tech import MailMicroService

env = {
    "SMTP_SERVER": "mail.test.com:587",
    "SMTP_LOGIN": "myself@test.com",
    "SMTP_PASSWD": "passwd"
}


@pytest.mark.local
class ATestMail:
    def test_send(self, local_server_factory, smtp_mock_fixture, email_mock_fixture):
        local_server = local_server_factory(MailMicroService(env=env))

        text_attachment = open("attachment.txt", "w+")
        text_attachment.write("This is the content of the attachemnt")
        text_attachment.close()
        text_attachment = open("attachment.txt", "r")

        response = local_server.make_call(requests.post, '/send',
                                          data={'subject': "Test mail",
                                                'from_addr': "myself@test.com",
                                                'to_addrs': ["you@test.com", "and@test.com"],
                                                'body': "content"},
                                          files={
                                              'attachments': ('attachment.txt', text_attachment, 'text/plain'),
                                          })
        text_attachment.close()
        os.remove("attachment.txt")
        assert response.status_code == 200
        smtp_mock_fixture.login.assert_called_once_with("myself@test.com", "passwd")
        email_mock_fixture.set_content.assert_called_once_with('content')
        email_mock_fixture.add_attachment.assert_called_with(b'This is the content of the attachemnt',
                                                             filename='attachment.txt', maintype='text', subtype='plain')
        assert email_mock_fixture["Subject"] == "Test mail"
        assert email_mock_fixture["From"] == "myself@test.com"
        assert email_mock_fixture["To"] == "you@test.com, and@test.com"
        assert response.text == "Mail sent to you@test.com, and@test.com"
        smtp_mock_fixture.send_message.assert_called()



import json

import requests

from coworks.blueprint import Admin
from tests.src.coworks.tech_ms import *


class DoumentedMS(TechMS):

    def get(self):
        """Root access."""
        return "get"

    def post_content(self, value, other="none"):
        """Add content."""
        return f"post_content {value}{other}"

    def post_contentannotated(self, value: int, other: str = "none"):
        """Add content."""
        return f"post_content {value}{other}"


def test_documentation(local_server_factory):
    ms = DoumentedMS()
    ms.register_blueprint(Admin(), url_prefix="/admin")
    local_server = local_server_factory(ms)
    response = local_server.make_call(requests.get, '/admin/routes')
    assert response.status_code == 200
    assert json.loads(response.text)["/"] == {
        "GET": {
            "doc": "Root access.",
            "signature": "()"
        }
    }
    assert json.loads(response.text)["/content/{_0}"] == {
        "POST": {
            "doc": "Add content.",
            "signature": "(value, other=\'none\')"
        }
    }
    assert json.loads(response.text)["/contentannotated/{_0}"] == {
        "POST": {
            "doc": "Add content.",
            "signature": "(value:int, other:str=\'none\')"
        }
    }
    assert json.loads(response.text)["/admin/routes"] == {
        "GET": {
            "doc": 'Returns the list of entrypoints with signature.',
            "signature": "()"
        }
    }



import requests
from chalice import AuthResponse

from .tech_ms import SimpleMS
from .blueprint.blueprint import BP


class AuthorizeAllMS(SimpleMS):

    def __init__(self):
        super().__init__()
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def auth(self, auth_request):
        return True


class AuthorizeNothingMS(SimpleMS):

    def __init__(self):
        super().__init__()
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def auth(self, auth_request):
        return False


class AuthorizedMS(SimpleMS):

    def auth(self, auth_request):
        token = auth_request.token
        if token == 'allow':
            return AuthResponse(routes=['/'], principal_id='user')
        return False


def test_authorize_all(local_server_factory):
    local_server = local_server_factory(AuthorizeAllMS())
    response = local_server.make_call(requests.get, '/', headers={'authorization': 'token'})
    assert response.status_code == 200
    assert response.text == 'get'
    response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'token'})
    assert response.status_code == 200
    assert response.text == 'blueprint test 3'


def test_authorize_nothing(local_server_factory):
    local_server = local_server_factory(AuthorizeNothingMS())
    response = local_server.make_call(requests.get, '/', headers={'authorization': 'token'})
    assert response.status_code == 403
    response = local_server.make_call(requests.get, '/blueprint/test/3', headers={'authorization': 'token'})
    assert response.status_code == 403


def test_authorized(local_server_factory):
    local_server = local_server_factory(AuthorizedMS())
    response = local_server.make_call(requests.get, '/', headers={'authorization': 'allow'})
    assert response.status_code == 200
    assert response.text == 'get'
    response = local_server.make_call(requests.get, '/', headers={'authorization': 'refuse'})
    assert response.status_code == 403
